from itertools import combinations

from ..core import Basic
from .graycode import GrayCode


class Subset(Basic):
    """
    Represents a basic subset object.

    We generate subsets using essentially two techniques,
    binary enumeration and lexicographic enumeration.
    The Subset class takes two arguments, the first one
    describes the initial subset to consider and the second
    describes the superset.

    Examples
    ========

    >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
    >>> a.next_binary().subset
    ['b']
    >>> a.prev_binary().subset
    ['c']

    """

    _rank_binary = None
    _rank_lex = None
    _rank_graycode = None
    _subset = None
    _superset = None

    def __new__(cls, subset, superset):
        """
        Default constructor.

        It takes the subset and its superset as its parameters.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.subset
        ['c', 'd']
        >>> a.superset
        ['a', 'b', 'c', 'd']
        >>> a.size
        2

        """
        if len(subset) > len(superset):
            raise ValueError('Invalid arguments have been provided. The '
                             'superset must be larger than the subset.')
        for elem in subset:
            if elem not in superset:
                raise ValueError('The superset provided is invalid as it '
                                 f'does not contain the element {elem}')
        obj = Basic.__new__(cls)
        obj._subset = subset
        obj._superset = superset
        return obj

    def iterate_binary(self, k):
        """
        This is a helper function. It iterates over the
        binary subsets by k steps. This variable can be
        both positive or negative.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.iterate_binary(-2).subset
        ['d']
        >>> a = Subset(['a', 'b', 'c'], ['a', 'b', 'c', 'd'])
        >>> a.iterate_binary(2).subset
        []

        See Also
        ========
        next_binary, prev_binary

        """
        bin_list = Subset.bitlist_from_subset(self.subset, self.superset)
        n = (int(''.join(bin_list), 2) + k) % 2**self.superset_size
        bits = bin(n)[2:].rjust(self.superset_size, '0')
        return Subset.subset_from_bitlist(self.superset, bits)

    def next_binary(self):
        """
        Generates the next binary ordered subset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.next_binary().subset
        ['b']
        >>> a = Subset(['a', 'b', 'c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.next_binary().subset
        []

        See Also
        ========
        prev_binary, iterate_binary

        """
        return self.iterate_binary(1)

    def prev_binary(self):
        """
        Generates the previous binary ordered subset.

        Examples
        ========

        >>> a = Subset([], ['a', 'b', 'c', 'd'])
        >>> a.prev_binary().subset
        ['a', 'b', 'c', 'd']
        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.prev_binary().subset
        ['c']

        See Also
        ========
        next_binary, iterate_binary

        """
        return self.iterate_binary(-1)

    def next_lexicographic(self):
        """
        Generates the next lexicographically ordered subset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.next_lexicographic().subset
        ['d']
        >>> a = Subset(['d'], ['a', 'b', 'c', 'd'])
        >>> a.next_lexicographic().subset
        []

        See Also
        ========
        prev_lexicographic

        """
        i = self.superset_size - 1
        indices = Subset.subset_indices(self.subset, self.superset)

        if i in indices:
            if i - 1 in indices:
                indices.remove(i - 1)
            else:
                indices.remove(i)
                i = i - 1
                while i not in indices and i >= 0:
                    i = i - 1
                if i >= 0:
                    indices.remove(i)
                    indices.append(i+1)
        else:
            while i not in indices and i >= 0:
                i = i - 1
            indices.append(i + 1)

        ret_set = []
        super_set = self.superset
        for i in indices:
            ret_set.append(super_set[i])
        return Subset(ret_set, super_set)

    def prev_lexicographic(self):
        """
        Generates the previous lexicographically ordered subset.

        Examples
        ========

        >>> a = Subset([], ['a', 'b', 'c', 'd'])
        >>> a.prev_lexicographic().subset
        ['d']
        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.prev_lexicographic().subset
        ['c']

        See Also
        ========
        next_lexicographic

        """
        i = self.superset_size - 1
        indices = Subset.subset_indices(self.subset, self.superset)

        while i not in indices and i >= 0:
            i = i - 1

        if i - 1 in indices or i == 0:
            indices.remove(i)
        else:
            if i >= 0:
                indices.remove(i)
                indices.append(i - 1)
            indices.append(self.superset_size - 1)

        ret_set = []
        super_set = self.superset
        for i in indices:
            ret_set.append(super_set[i])
        return Subset(ret_set, super_set)

    def iterate_graycode(self, k):
        """
        Helper function used for prev_gray and next_gray.
        It performs k step overs to get the respective Gray codes.

        Examples
        ========

        >>> a = Subset([1, 2, 3], [1, 2, 3, 4])
        >>> a.iterate_graycode(3).subset
        [1, 4]
        >>> a.iterate_graycode(-2).subset
        [1, 2, 4]

        See Also
        ========
        next_gray, prev_gray

        """
        unranked_code = GrayCode.unrank(self.superset_size,
                                        (self.rank_gray + k) % self.cardinality)
        return Subset.subset_from_bitlist(self.superset,
                                          unranked_code)

    def next_gray(self):
        """
        Generates the next Gray code ordered subset.

        Examples
        ========

        >>> a = Subset([1, 2, 3], [1, 2, 3, 4])
        >>> a.next_gray().subset
        [1, 3]

        See Also
        ========
        iterate_graycode, prev_gray

        """
        return self.iterate_graycode(1)

    def prev_gray(self):
        """
        Generates the previous Gray code ordered subset.

        Examples
        ========

        >>> a = Subset([2, 3, 4], [1, 2, 3, 4, 5])
        >>> a.prev_gray().subset
        [2, 3, 4, 5]

        See Also
        ========
        iterate_graycode, next_gray

        """
        return self.iterate_graycode(-1)

    @property
    def rank_binary(self):
        """
        Computes the binary ordered rank.

        Examples
        ========

        >>> a = Subset([], ['a', 'b', 'c', 'd'])
        >>> a.rank_binary
        0
        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.rank_binary
        3

        See Also
        ========
        iterate_binary, unrank_binary

        """
        if self._rank_binary is None:
            self._rank_binary = int(''.join(
                Subset.bitlist_from_subset(self.subset,
                                           self.superset)), 2)
        return self._rank_binary

    @property
    def rank_lexicographic(self):
        """
        Computes the lexicographic ranking of the subset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.rank_lexicographic
        14
        >>> a = Subset([2, 4, 5], [1, 2, 3, 4, 5, 6])
        >>> a.rank_lexicographic
        43

        """
        if self._rank_lex is None:
            def _ranklex(self, subset_index, i, n):
                if subset_index == [] or i > n:
                    return 0
                if i in subset_index:
                    subset_index.remove(i)
                    return 1 + _ranklex(self, subset_index, i + 1, n)
                return 2**(n - i - 1) + _ranklex(self, subset_index, i + 1, n)
            indices = Subset.subset_indices(self.subset, self.superset)
            self._rank_lex = _ranklex(self, indices, 0, self.superset_size)
        return self._rank_lex

    @property
    def rank_gray(self):
        """
        Computes the Gray code ranking of the subset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.rank_gray
        2
        >>> a = Subset([2, 4, 5], [1, 2, 3, 4, 5, 6])
        >>> a.rank_gray
        27

        See Also
        ========
        iterate_graycode, unrank_gray

        """
        if self._rank_graycode is None:
            bits = Subset.bitlist_from_subset(self.subset, self.superset)
            self._rank_graycode = GrayCode(len(bits), start=bits).rank
        return self._rank_graycode

    @property
    def subset(self):
        """
        Gets the subset represented by the current instance.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.subset
        ['c', 'd']

        See Also
        ========
        superset, size, superset_size, cardinality

        """
        return self._subset

    @property
    def size(self):
        """
        Gets the size of the subset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.size
        2

        See Also
        ========
        subset, superset, superset_size, cardinality

        """
        return len(self.subset)

    @property
    def superset(self):
        """
        Gets the superset of the subset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.superset
        ['a', 'b', 'c', 'd']

        See Also
        ========
        subset, size, superset_size, cardinality

        """
        return self._superset

    @property
    def superset_size(self):
        """
        Returns the size of the superset.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.superset_size
        4

        See Also
        ========
        subset, superset, size, cardinality

        """
        return len(self.superset)

    @property
    def cardinality(self):
        """
        Returns the number of all possible subsets.

        Examples
        ========

        >>> a = Subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        >>> a.cardinality
        16

        See Also
        ========
        subset, superset, size, superset_size

        """
        return 2**(self.superset_size)

    @classmethod
    def subset_from_bitlist(cls, super_set, bitlist):
        """
        Gets the subset defined by the bitlist.

        Examples
        ========

        >>> Subset.subset_from_bitlist(['a', 'b', 'c', 'd'], '0011').subset
        ['c', 'd']

        See Also
        ========
        bitlist_from_subset

        """
        if len(super_set) != len(bitlist):
            raise ValueError('The sizes of the lists are not equal')
        ret_set = []
        for i, bi in enumerate(bitlist):
            if bi == '1':
                ret_set.append(super_set[i])
        return Subset(ret_set, super_set)

    @classmethod
    def bitlist_from_subset(cls, subset, superset):
        """
        Gets the bitlist corresponding to a subset.

        Examples
        ========

        >>> Subset.bitlist_from_subset(['c', 'd'], ['a', 'b', 'c', 'd'])
        '0011'

        See Also
        ========
        subset_from_bitlist

        """
        bitlist = ['0'] * len(superset)
        for i in Subset.subset_indices(subset, superset):
            bitlist[i] = '1'
        return ''.join(bitlist)

    @classmethod
    def unrank_binary(cls, rank, superset):
        """
        Gets the binary ordered subset of the specified rank.

        Examples
        ========

        >>> Subset.unrank_binary(4, ['a', 'b', 'c', 'd']).subset
        ['b']

        See Also
        ========
        iterate_binary, rank_binary

        """
        bits = bin(rank)[2:].rjust(len(superset), '0')
        return Subset.subset_from_bitlist(superset, bits)

    @classmethod
    def unrank_gray(cls, rank, superset):
        """
        Gets the Gray code ordered subset of the specified rank.

        Examples
        ========

        >>> Subset.unrank_gray(4, ['a', 'b', 'c']).subset
        ['a', 'b']
        >>> Subset.unrank_gray(0, ['a', 'b', 'c']).subset
        []

        See Also
        ========
        iterate_graycode, rank_gray

        """
        graycode_bitlist = GrayCode.unrank(len(superset), rank)
        return Subset.subset_from_bitlist(superset, graycode_bitlist)

    @classmethod
    def subset_indices(cls, subset, superset):
        """Return indices of subset in superset in a list; the list is empty
        if all elements of subset are not in superset.

        Examples
        ========

            >>> superset = [1, 3, 2, 5, 4]
            >>> Subset.subset_indices([3, 2, 1], superset)
            [1, 2, 0]
            >>> Subset.subset_indices([1, 6], superset)
            []
            >>> Subset.subset_indices([], superset)
            []

        """
        a, b = superset, subset
        sb = set(b)
        d = {}
        for i, ai in enumerate(a):
            if ai in sb:
                d[ai] = i
                sb.remove(ai)
                if not sb:
                    break
        else:
            return []
        return [d[bi] for bi in b]


def ksubsets(superset, k):
    """
    Finds the subsets of size k in lexicographic order.

    This uses the itertools generator.

    Examples
    ========

    >>> list(ksubsets([1, 2, 3], 2))
    [(1, 2), (1, 3), (2, 3)]
    >>> list(ksubsets([1, 2, 3, 4, 5], 2))
    [(1, 2), (1, 3), (1, 4), (1, 5), (2, 3), (2, 4),
     (2, 5), (3, 4), (3, 5), (4, 5)]

    See Also
    ========

    Subset

    """
    return combinations(superset, k)



"""Limit computation tests."""

import itertools

import pytest

from diofant import (E, Float, Function, I, Integral, Lambda, Limit, O,
                     Piecewise, PoleError, Rational, Reals, RootSum, Sum,
                     Symbol, acos, acosh, acoth, asin, atan, besselk, binomial,
                     cbrt, ceiling, cos, cosh, cot, diff, digamma, erf, erfc,
                     erfi, exp, factorial, false, floor, gamma, integrate,
                     limit, log, nan, oo, pi, polygamma, root, sign, simplify,
                     sin, sinh, sqrt, subfactorial, symbols, tan, true)
from diofant.abc import a, b, c, k, n, x, y, z
from diofant.series.limits import heuristics


__all__ = ()


def test_basic1():
    assert limit(x, x, -oo) == -oo
    assert limit(x**2, x, -oo) == oo
    assert limit(x*log(x), x, 0) == 0
    assert limit(x - x**2, x, oo) == -oo
    assert limit((1 + x)**(1 + sqrt(2)), x, 0) == 1
    assert limit((1 + x)**oo, x, 0) == oo
    assert limit((1 + x)**oo, x, 0, dir=1) == 0
    assert limit((1 + x + y)**oo, x, 0, dir=1) == (1 + y)**oo
    assert limit(y/x/log(x), x, 0) == -oo*sign(y)
    assert limit(cos(x + y)/x, x, 0) == sign(cos(y))*oo
    limit(Sum(1/x, (x, 1, y)) - log(y), y, oo)
    limit(Sum(1/x, (x, 1, y)) - 1/y, y, oo)
    assert limit(nan, x, -oo) == nan
    assert limit(O(2)*x, x, nan) == nan
    assert limit(sin(O(x)), x, 0) == 0
    assert limit(1/(x - 1), x, 1) == oo
    assert limit(1/(x - 1), x, 1, dir=1) == -oo
    assert limit(1/(5 - x)**3, x, 5) == -oo
    assert limit(1/(5 - x)**3, x, 5, dir=1) == oo
    assert limit(1/sin(x), x, pi) == -oo
    assert limit(1/sin(x), x, pi, dir=1) == oo
    assert limit(1/cos(x), x, pi/2) == -oo
    assert limit(1/cos(x), x, pi/2, dir=1) == oo
    assert limit(1/tan(x**3), x, cbrt(2*pi)) == oo
    assert limit(1/tan(x**3), x, cbrt(2*pi), dir=1) == -oo
    assert limit(1/cot(x)**3, x, 3*pi/2) == -oo
    assert limit(1/cot(x)**3, x, 3*pi/2, dir=1) == oo

    # approaching 0
    # from dir=-1
    assert limit(1 + 1/x, x, 0) == oo
    # from dir=1
    # Add
    assert limit(1 + 1/x, x, 0, dir=1) == -oo
    # Pow
    assert limit(x**(-2), x, 0, dir=1) == oo
    assert limit(x**(-3), x, 0, dir=1) == -oo
    assert limit(1/sqrt(x), x, 0, dir=1) == (-oo)*I
    assert limit(x**2, x, 0, dir=1) == 0
    assert limit(sqrt(x), x, 0, dir=1) == 0
    assert limit(x**-pi, x, 0, dir=1) == oo*sign((-1)**(-pi))
    assert limit((1 + cos(x))**oo, x, 0) == oo

    assert limit(x**2, x, 0, dir=Reals) == 0
    assert limit(exp(x), x, 0, dir=Reals) == 1

    pytest.raises(PoleError, lambda: limit(1/x, x, 0, dir=Reals))

    # issue diofant/diofant#74
    assert limit(sign(log(1 - 1/x)), x, oo) == -1

    # issue sympy/sympy#8166
    f = Function('f')
    assert limit(f(x), x, 4) == Limit(f(x), x, 4)

    assert limit(exp(x), x, 0, dir=exp(I*pi/3)) == 1

    assert limit(sqrt(-1 + I*x), x, 0) == +I
    assert limit(sqrt(-1 + I*x), x, 0, dir=1) == -I
    assert limit(sqrt(-1 + I*x), x, 0, dir=exp(I*pi/3)) == -I

    assert limit(log(x + sqrt(x**2 + 1)), x, I) == I*pi/2
    assert limit(log(x + sqrt(x**2 + 1)), x, I, dir=1) == I*pi/2
    assert limit(log(x + sqrt(x**2 + 1)), x, I, dir=exp(I*pi/3)) == I*pi/2


def test_basic2():
    assert limit(x**x, x, 0) == 1
    assert limit((exp(x) - 1)/x, x, 0) == 1


def test_basic3():
    assert limit(1/x, x, 0) == oo
    assert limit(1/x, x, 0, dir=1) == -oo


def test_basic4():
    assert limit(2*x + y*x, x, 0) == 0
    assert limit(2*x + y*x, x, 1) == 2 + y
    assert limit(2*x**8 + y*x**(-3), x, -2) == 512 - y/8
    assert limit(sqrt(x + 1) - sqrt(x), x, oo) == 0
    assert integrate(1/(x**3 + 1), (x, 0, oo)) == 2*pi*sqrt(3)/9

    # coverage test
    assert limit(Piecewise((x, x > 1), (0, True)), x, -1) == 0

    # issue sympy/sympy#16714
    e = ((n**(n + 1) + (n + 1)**n)/n**(n + 1))**n
    assert limit(e, n, oo) == E**E

    # issue sympy/sympy#18492
    e1 = 2*sqrt(x)*Piecewise(((4*x - 2)/abs(sqrt(4 - 4*(2*x - 1)**2)),
                              4*x - 2 >= 0),
                             ((2 - 4*x)/abs(sqrt(4 - 4*(2*x - 1)**2)), True))
    e2 = Piecewise((x**2/2, x <= Rational(1, 2)), (x/2 - Rational(1, 8), True))
    e3 = Piecewise(((x - 9)/5, x < -1), ((x - 9)/5, x > 4),
                   (sqrt(abs(x - 3)), True))
    assert limit(e1, x, 0) == 1
    assert limit(e2, x, 0) == 0
    assert limit(e2, x, oo) == oo
    assert limit(e3, x, -1) == 2
    assert limit(e3, x, oo) == oo

    e4 = Piecewise((1, 0 < x), (0, True))

    assert limit(e4, x, 0, 1) == 0
    assert limit(e4, x, 0) == 1
    pytest.raises(PoleError, lambda: limit(e4, x, 0, Reals))

    e5 = Piecewise((1, 0 < x), (2, 1 < x), (0, True))

    assert limit(e5, x, oo) == 1
    assert limit(e5, x, 1, 1) == 1
    assert limit(e5, x, 1) == 1

    e6 = Limit(Piecewise((1, x > a), (0, True)), x, 0)
    assert e6.doit() == e6


def test_basic5():
    class MyFunction(Function):
        @classmethod
        def eval(cls, arg):
            if arg is oo:
                return nan
    assert limit(MyFunction(x), x, oo) == Limit(MyFunction(x), x, oo)

    assert limit(4/x > 8, x, 0) is true  # relational test
    assert limit(MyFunction(x) > 0, x, oo) == Limit(MyFunction(x) > 0, x, oo)

    # issue diofant/diofant#1217
    assert limit(x > 0, x, 0) is true
    assert limit(x > 0, x, 0, 1) is false

    # issue sympy/sympy#11833
    a = Symbol('a', positive=True)
    f = exp(x*(-a - 1)) * sinh(x)
    assert limit(f, x, oo) == 0

    assert limit(O(x), x, x**2) == Limit(O(x), x, x**2)


def test_basic6():
    # pull sympy/sympy#22491
    assert limit(1/asin(x), x, 0) == oo
    assert limit(1/asin(x), x, 0, dir=1) == -oo
    assert limit(1/sinh(x), x, 0) == oo
    assert limit(1/sinh(x), x, 0, dir=1) == -oo
    assert limit(log(1/x) + 1/sin(x), x, 0) == oo
    assert limit(log(1/x) + 1/x, x, 0) == oo


def test_sympyissue_3885():
    assert limit(x*y + x*z, z, 2) == x*y + 2*x


def test_Limit():
    assert Limit(sin(x)/x, x, 0) != 1
    assert Limit(sin(x)/x, x, 0).doit() == 1


def test_floor():
    assert limit(floor(x), x, -2) == -2
    assert limit(floor(x), x, -2, 1) == -3
    assert limit(floor(x), x, -1) == -1
    assert limit(floor(x), x, -1, 1) == -2
    assert limit(floor(x), x, 0) == 0
    assert limit(floor(x), x, 0, 1) == -1
    assert limit(floor(x), x, 1) == 1
    assert limit(floor(x), x, 1, 1) == 0
    assert limit(floor(x), x, 2) == 2
    assert limit(floor(x), x, 2, 1) == 1
    assert limit(floor(x), x, 248) == 248
    assert limit(floor(x), x, 248, 1) == 247


def test_floor_requires_robust_assumptions():
    assert limit(floor(sin(x)), x, 0) == 0
    assert limit(floor(sin(x)), x, 0, 1) == -1
    assert limit(floor(cos(x)), x, 0) == 0
    assert limit(floor(cos(x)), x, 0, 1) == 0
    assert limit(floor(5 + sin(x)), x, 0) == 5
    assert limit(floor(5 + sin(x)), x, 0, 1) == 4
    assert limit(floor(5 + cos(x)), x, 0) == 5
    assert limit(floor(5 + cos(x)), x, 0, 1) == 5


def test_ceiling():
    assert limit(ceiling(x), x, -2) == -1
    assert limit(ceiling(x), x, -2, 1) == -2
    assert limit(ceiling(x), x, -1) == 0
    assert limit(ceiling(x), x, -1, 1) == -1
    assert limit(ceiling(x), x, 0) == 1
    assert limit(ceiling(x), x, 0, 1) == 0
    assert limit(ceiling(x), x, 1) == 2
    assert limit(ceiling(x), x, 1, 1) == 1
    assert limit(ceiling(x), x, 2) == 3
    assert limit(ceiling(x), x, 2, 1) == 2
    assert limit(ceiling(x), x, 248) == 249
    assert limit(ceiling(x), x, 248, 1) == 248


def test_ceiling_requires_robust_assumptions():
    assert limit(ceiling(sin(x)), x, 0) == 1
    assert limit(ceiling(sin(x)), x, 0, 1) == 0
    assert limit(ceiling(cos(x)), x, 0) == 1
    assert limit(ceiling(cos(x)), x, 0, 1) == 1
    assert limit(ceiling(5 + sin(x)), x, 0) == 6
    assert limit(ceiling(5 + sin(x)), x, 0, 1) == 5
    assert limit(ceiling(5 + cos(x)), x, 0) == 6
    assert limit(ceiling(5 + cos(x)), x, 0, 1) == 6


def test_atan():
    assert limit(atan(x)*sin(1/x), x, 0) == 0
    assert limit(atan(x) + sqrt(x + 1) - sqrt(x), x, oo) == pi/2


def test_acosh():
    assert limit(acosh(I*x), x, 0) == 2*log(1 + I) - log(2)
    assert limit(acosh(I*x), x, 0, 1) == 2*log(1 - I) - log(2)
    assert limit(acosh(-I*x), x, 0) == 2*log(1 - I) - log(2)
    assert limit(acosh(-I*x), x, 0, 1) == 2*log(1 + I) - log(2)
    pytest.raises(PoleError, lambda: limit(acosh(I*x), x, 0, Reals))
    assert limit(acosh(x - I*x), x, 0) == 2*log(1 - I) - log(2)
    assert limit(acosh(x - I*x), x, 0, 1) == 2*log(1 + I) - log(2)
    pytest.raises(PoleError, lambda: limit(acosh(x - I*x), x, 0, Reals))
    assert limit(acosh(I*x**2), x, 0, Reals) == 2*log(1 + I) - log(2)
    assert limit(acosh(x**2 - I*x**2), x, 0, Reals) == 2*log(1 - I) - log(2)


def test_acoth():
    assert limit(acoth(x), x, 0) == -I*pi/2
    assert limit(acoth(x), x, 0, 1) == I*pi/2
    pytest.raises(PoleError, lambda: limit(acoth(x), x, 0, Reals))
    assert limit(acoth(-x), x, 0) == I*pi/2
    assert limit(acoth(x**2), x, 0, Reals) == -I*pi/2
    assert limit(acoth(I*x), x, 0) == -I*pi/2
    assert limit(acoth(I*x), x, 0, 1) == I*pi/2
    pytest.raises(PoleError, lambda: limit(acoth(x - I*x), x, 0, Reals))


def test_abs():
    assert limit(abs(x), x, 0) == 0
    assert limit(abs(sin(x)), x, 0) == 0
    assert limit(abs(cos(x)), x, 0) == 1
    assert limit(abs(sin(x + 1)), x, 0) == sin(1)

    # sympy/sympy#12398
    assert limit(abs(log(n)/n**3), n, oo) == 0
    expr = abs(log(n)/n**3)
    expr2 = expr.subs({n: n + 1})
    assert limit(n*(expr/expr2 - 1), n, oo) == 3


def test_heuristic():
    x = Symbol('x', extended_real=True)
    assert heuristics(sin(1/x) + atan(x), x, 0, -1) == sin(oo)
    assert heuristics(log(2 + sqrt(atan(x))*sin(1/x)), x, 0, -1) == log(2)
    assert heuristics(tan(tan(1/x)), x, 0, -1) is None


def test_sympyissue_3871():
    z = Symbol('z', positive=True)
    f = -1/z*exp(-z*x)
    assert limit(f, x, oo) == 0
    assert f.limit(x, oo) == 0


def test_exponential():
    x = Symbol('x', extended_real=True)
    assert limit((1 + x/n)**n, n, oo) == exp(x)
    assert limit((1 + x/(2*n))**n, n, oo) == exp(x/2)
    assert limit((1 + x/(2*n + 1))**n, n, oo) == exp(x/2)
    assert limit(((x - 1)/(x + 1))**x, x, oo) == exp(-2)
    assert limit(1 + (1 + 1/x)**x, x, oo) == 1 + E


@pytest.mark.xfail
def test_exponential2():
    assert limit((1 + x/(n + sin(n)))**n, n, oo) == exp(x)


def test_doit():
    f = Integral(2 * x, x)
    l = Limit(f, x, oo)
    assert l.doit() == oo


def test_doit2():
    f = Integral(2 * x, x)
    l = Limit(f, x, oo)
    # limit() breaks on the contained Integral.
    assert l.doit(deep=False) == l


def test_RootSum_limits():
    r = RootSum(x**2 + 1, Lambda(x, x*log(x + y)))

    assert limit(r, y, oo) == 0


def test_sympyissue_3792():
    assert limit((1 - cos(x))/x**2, x, Rational(1, 2)) == 4 - 4*cos(Rational(1, 2))
    assert limit(sin(sin(x + 1) + 1), x, 0) == sin(1 + sin(1))
    assert limit(abs(sin(x + 1) + 1), x, 0) == 1 + sin(1)


def test_sympyissue_4090():
    assert limit(1/(x + 3), x, 2) == Rational(1, 5)
    assert limit(1/(x + pi), x, 2) == 1/(2 + pi)
    assert limit(log(x)/(x**2 + 3), x, 2) == log(2)/7
    assert limit(log(x)/(x**2 + pi), x, 2) == log(2)/(4 + pi)


def test_sympyissue_4547():
    assert limit(cot(x), x, 0) == oo
    assert limit(cot(x), x, pi/2) == 0


def test_sympyissue_5164():
    assert limit(x**0.5, x, oo) == oo**0.5 == oo
    assert limit(x**0.5, x, 16) == 4.0
    assert limit(x**0.5, x, 0) == 0
    assert limit(x**(-0.5), x, oo) == 0
    assert limit(x**(-0.5), x, 4) == 0.5


def test_sympyissue_5183():
    # using list(...) so pytest can recalculate values
    tests = list(itertools.product([x, -x],
                                   [-1, 1],
                                   [2, 3, Rational(1, 2), Rational(2, 3)],
                                   [1, -1]))
    results = (oo, oo, -oo, oo, -oo*I, oo, -oo*cbrt(-1), oo,
               0, 0, 0, 0, 0, 0, 0, 0,
               oo, oo, oo, -oo, oo, -oo*I, oo, -oo*cbrt(-1),
               0, 0, 0, 0, 0, 0, 0, 0)
    assert len(tests) == len(results)
    for args, res in zip(tests, results):
        y, s, e, d = args
        eq = y**(s*e)
        assert limit(eq, x, 0, dir=d) == res


def test_sympyissue_5184():
    assert limit(sin(x)/x, x, oo) == 0
    assert limit(atan(x), x, oo) == pi/2
    assert limit(gamma(x), x, oo) == oo
    assert limit(cos(x)/x, x, oo) == 0
    assert limit(gamma(x), x, Rational(1, 2)) == sqrt(pi)

    r = Symbol('r', real=True)
    assert limit(r*sin(1/r), r, 0) == 0


def test_sympyissue_5229():
    assert limit((1 + y)**(1/y) - E, y, 0) == 0


def test_sympyissue_4546():
    # using list(...) so pytest can recalculate values
    tests = list(itertools.product([cot, tan],
                                   [-pi/2, 0, pi/2, pi, 3*pi/2],
                                   [1, -1]))
    results = (0, 0, -oo, oo, 0, 0, -oo, oo, 0, 0,
               oo, -oo, 0, 0, oo, -oo, 0, 0, oo, -oo)
    assert len(tests) == len(results)
    for args, res in zip(tests, results):
        f, l, d = args
        eq = f(x)
        assert limit(eq, x, l, dir=d) == res


def test_sympyissue_3934():
    assert limit((1 + x**log(3))**(1/x), x, 0) == 1
    assert limit((5**(1/x) + 3**(1/x))**x, x, 0) == 5


def test_compute_leading_term():
    assert limit(x**Rational(77, 3)/(1 + x**Rational(77, 3)), x, oo) == 1
    assert limit(x**Rational('101.1')/(1 + x**Rational('101.1')), x, oo) == 1


def test_sympyissue_5955():
    assert limit((x**16)/(1 + x**16), x, oo) == 1
    assert limit((x**100)/(1 + x**100), x, oo) == 1
    assert limit((x**1885)/(1 + x**1885), x, oo) == 1
    assert limit((x**100/((x + 1)**100 + exp(-x))), x, oo) == 1


def test_newissue():
    assert limit(exp(1/sin(x))/exp(cot(x)), x, 0) == 1


def test_extended_real_line():
    assert limit(x - oo, x, oo) == -oo
    assert limit(oo - x, x, -oo) == oo
    assert limit(x**2/(x - 5) - oo, x, oo) == -oo
    assert limit(1/(x + sin(x)) - oo, x, 0) == -oo
    assert limit(oo/x, x, oo) == oo
    assert limit(x - oo + 1/x, x, oo) == -oo
    assert limit(x - oo + 1/x, x, 0) == -oo


@pytest.mark.xfail
def test_order_oo():
    x = Symbol('x', positive=True, finite=True)
    assert O(x)*oo != O(1, x)
    assert limit(oo/(x**2 - 4), x, oo) == oo


def test_sympyissue_5436():
    # also issue sympy/sympy#13312 (but see diofant/diofant#425!)
    assert limit(exp(x*y), x, oo) == exp(oo*sign(y))
    assert limit(exp(-x*y), x, oo) == exp(-oo*sign(y))


def test_Limit_dir():
    pytest.raises(ValueError, lambda: Limit(x, x, 0, dir=0))
    pytest.raises(ValueError, lambda: Limit(x, x, 0, dir='0'))


def test_polynomial():
    assert limit((x + 1)**1000/((x + 1)**1000 + 1), x, oo) == 1
    assert limit((x + 1)**1000/((x + 1)**1000 + 1), x, -oo) == 1


def test_rational():
    assert limit(1/y - (1/(y + x) + x/(y + x)/y)/z, x, oo) == (z - 1)/(y*z)
    assert limit(1/y - (1/(y + x) + x/(y + x)/y)/z, x, -oo) == (z - 1)/(y*z)


def test_sympyissue_5740():
    assert limit(log(x)*z - log(2*x)*y, x, 0) == oo*sign(y - z)


def test_sympyissue_6366():
    n = Symbol('n', integer=True, positive=True)
    r = (n + 1)*x**(n + 1)/(x**(n + 1) - 1) - x/(x - 1)
    assert limit(r, x, 1).simplify() == n/2


def test_factorial():
    f = factorial(x)
    assert limit(f, x, oo) == oo
    assert limit(x/f, x, oo) == 0
    # see Stirling's approximation:
    # https://en.wikipedia.org/wiki/Stirling's_approximation
    assert limit(f/(sqrt(2*pi*x)*(x/E)**x), x, oo) == 1
    assert limit(f, x, -oo) == factorial(-oo)
    assert (limit(f, x, x**2) - factorial(x**2)).simplify() == 0
    assert (limit(f, x, -x**2) - factorial(-x**2)).simplify() == 0


def test_sympyissue_6560():
    e = 5*x**3/4 - 3*x/4 + (y*(3*x**2/2 - Rational(1, 2)) +
                            35*x**4/8 - 15*x**2/4 + Rational(3, 8))/(2*(y + 1))
    assert limit(e, y, oo) == (5*x**3 + 3*x**2 - 3*x - 1)/4


def test_sympyissue_5172():
    r = Symbol('r', positive=True)
    p = Symbol('p', positive=True)
    m = Symbol('m', negative=True)
    expr = ((2*n*(n - r + 1)/(n + r*(n - r + 1)))**c +
            (r - 1)*(n*(n - r + 2)/(n + r*(n - r + 1)))**c - n)/(n**c - n)
    expr = expr.subs({c: c + 1})
    assert limit(expr.subs({c: m}), n, oo) == 1
    assert limit(expr.subs({c: p}), n, oo) == (2**(p + 1) + r -
                                               1)/(r + 1)**(p + 1)


def test_sympyissue_7088():
    assert limit(sqrt(x/(x + a)), x, oo) == 1


def test_sympyissue_6364():
    e = z/(1 - sqrt(1 + z)*sin(a)**2 - sqrt(1 - z)*cos(a)**2)
    assert (limit(e, z, 0) - 2/cos(2*a)).simplify() == 0


def test_sympyissue_4099():
    assert limit(a/x, x, 0) == oo*sign(a)
    assert limit(-a/x, x, 0) == -oo*sign(a)
    assert limit(-a*x, x, oo) == -oo*sign(a)
    assert limit(a*x, x, oo) == oo*sign(a)


def test_sympyissue_4503():
    assert limit((sqrt(1 + exp(x + y)) - sqrt(1 + exp(x)))/y,
                 y, 0) == exp(x)/(2*sqrt(exp(x) + 1))


def test_sympyissue_8730():
    assert limit(subfactorial(x), x, oo) == oo


def test_issue_55():
    assert limit((x + exp(x))/(x - 1), x, -oo) == 1
    assert limit((x*exp(x))/(exp(x) - 1), x, -oo) == 0  # issue sympy/sympy#2929

    # issue sympy/sympy#22982
    assert limit((log(E + 1/x) - 1)**(1 - sqrt(E + 1/x)), x, oo) == oo
    assert limit((log(E + 1/x))**(1 - sqrt(E + 1/x)), x, oo) == 1
    assert limit((log(E + 1/x) - 1)**-sqrt(E + 1/x), x, oo) == oo


def test_sympyissue_8061():
    assert limit(4**(acos(1/(1 + x**2))**2)/log(1 + x, 4), x, 0) == oo


def test_sympyissue_8229():
    assert limit((root(x, 4) - 2)/(sqrt(x) - 4)**Rational(2, 3),
                 x, 16) == 0


def test_Limit_free_symbols():
    # issue sympy/sympy#9205
    assert Limit(x, x, a).free_symbols == {a}
    assert Limit(x, x, a, 1).free_symbols == {a}
    assert Limit(x + y, x + y, a).free_symbols == {a}
    assert Limit(-x**2 + y, x**2, a).free_symbols == {y, a}


def test_sympyissue_10610():
    assert limit(3**x*3**(-x - 1)*(x + 1)**2/x**2, x, oo) == Rational(1, 3)
    assert limit(2**x*2**(-x - 1)*(x + 1)*(y - 1)**(-x) *
                 (y - 1)**(x + 1)/(x + 2), x, oo) == y/2 - Rational(1, 2)


def test_sympyissue_9075():
    assert limit((6**(x + 1) + x + 1)/(6**x + x), x, oo) == 6


def test_sympyissue_8634():
    p = Symbol('p', positive=True)
    assert limit(x**p, x, -oo) == oo*sign((-1)**p)


def test_sympyissue_9558():
    assert limit(sin(x)**15, x, 0, 1) == 0  # should be fast


def test_issue_296():
    e = log(exp(1/x)/Float(2) + exp(-1/x)/2)*x**2
    assert e.limit(x, oo) == 0.5


def test_sympyissue_5383():
    e = (1.0 + 1.0*x)**(1.0/x)
    assert e.limit(x, 0) == E.evalf()


def test_sympyissue_6171():
    e = Piecewise((0, x < 0), (1, True))
    assert e.limit(x, 0) == 1
    assert e.limit(x, 0, 1) == 0


def test_sympyissue_11526():
    df = diff(1/(a*log((x - b)/(x - c))), x)
    res = -1/(-a*c + a*b)
    assert limit(df, x, oo) == res
    assert (limit(simplify(df), x, oo) - res).simplify() == 0

    e = log((1/x - b)/(1/x - c))
    assert e.as_leading_term(x) == x*(c - b)


def test_sympyissue_11672():
    assert limit(Rational(-1, 2)**x, x, oo) == 0
    assert limit(1/(-2)**x, x, oo) == 0


def test_sympyissue_8635():
    n = Symbol('n', integer=True, positive=True)

    k = 0
    assert limit(x**n - x**(n - k), x, oo) == 0
    k = 1
    assert limit(x**n - x**(n - k), x, oo) == oo
    k = 2
    assert limit(x**n - x**(n - k), x, oo) == oo
    k = 3
    assert limit(x**n - x**(n - k), x, oo) == oo


def test_sympyissue_8157():
    n = Symbol('n', integer=True)
    limit(cos(pi*n), n, oo)  # not raises


def test_sympyissue_5415():
    assert limit(polygamma(2 + 1/x, 3 + exp(-x)), x, oo) == polygamma(2, 3)


def test_sympyissue_2865():
    l1 = limit(O(1/x, (x, oo)), x, 0)
    assert l1 != 0
    assert isinstance(l1, Limit)
    l2 = limit(O(x, (x, oo)), x, 0)
    assert l2 != 0
    assert isinstance(l2, Limit)


def test_sympyissue_11879():
    assert limit(((x + y)**n - x**n)/y, y, 0).powsimp() == n*x**(n-1)


def test_sympyissue_12555():
    assert limit((3**x + 2*x**10)/(x**10 + E**x), x, -oo) == 2


def test_sympyissue_12769():
    r = Symbol('r', real=True)
    a, b, s0, K, F0 = symbols('a b s0 K F0', positive=True, real=True)
    fx = (F0**b*K**b*r*s0 -
          sqrt((F0**2*K**(2*b)*a**2*(b - 1) +
                F0**(2*b)*K**2*a**2*(b - 1) +
                F0**(2*b)*K**(2*b)*s0**2*(b - 1)*(b**2 - 2*b + 1) -
                2*F0**(2*b)*K**(b + 1)*a*r*s0*(b**2 - 2*b + 1) +
                2*F0**(b + 1)*K**(2*b)*a*r*s0*(b**2 - 2*b + 1) -
                2*F0**(b + 1)*K**(b + 1)*a**2*(b - 1))/((b - 1)*(b**2 - 2*b + 1))))*(b*r - b - r + 1)
    assert limit(fx, K, F0) == (F0**(2*b)*b*r**2*s0 - 2*F0**(2*b)*b*r*s0 +
                                F0**(2*b)*b*s0 - F0**(2*b)*r**2*s0 +
                                2*F0**(2*b)*r*s0 - F0**(2*b)*s0)


def test_sympyissue_13332():
    assert limit(sqrt(30)*5**(-5*n - 1)*(46656*n)**n *
                 (5*n + 2)**(5*n + Rational(5, 2)) *
                 (6*n + 2)**(-6*n - Rational(5, 2)), n, oo) == Rational(25, 36)


def test_sympyissue_13382():
    assert limit(n*(((n + 1)**2 + 1)/(n**2 + 1) - 1), n, oo) == 2


def test_sympyissue_13403():
    assert limit(n*(-1 + (n + log(n + 1) + 1)/(n + log(n))), n, oo) == 1


def test_sympyissue_13416():
    assert limit((-n**3*log(n)**3 +
                  (n - 1)*(n + 1)**2*log(n + 1)**3)/(n**2*log(n)**3),
                 n, oo) == 1


def test_sympyissue_13462():
    assert limit(n**2*(2*n*(-(1 - 1/(2*n))**x + 1) -
                 x - (-x**2/4 + x/4)/n), n, oo) == x/12 - x**2/8 + x**3/24


def test_sympyissue_13575():
    assert limit(acos(erfi(x)), x, 1).equals(pi/2 + I*log(sqrt(erf(I)**2 + 1) +
                                                          erf(I))) is True


def test_issue_558():
    r = Symbol('r', positive=True)
    expr = ((2*n*(n - r + 1)/(n + r*(n - r + 1)))**c +
            (r - 1)*(n*(n - r + 2)/(n + r*(n - r + 1)))**c - n)/(n**c - n)
    expr = expr.subs({c: c + 1})
    assert limit(expr, n, oo) == Limit(expr, n, oo)


def test_sympyissue_14393():
    assert limit((x**b - y**b)/(x**a - y**a), x, y) == b*y**b/y**a/a


def test_sympyissue_14590():
    assert limit((n**3*((n + 1)/n)**n)/((n + 1)*(n + 2)*(n + 3)), n, oo) == E


def test_sympyissue_14793():
    e = ((x + Rational(1, 2))*log(x) - x +
         log(2*pi)/2 - log(factorial(x)) + 1/(12*x))*x**3
    assert limit(e, x, oo) == Rational(1, 360)


def test_sympyissue_14811():
    assert limit(((1 + Rational(2, 3)**(x + 1))**2**x)/(2**Rational(4, 3)**(x - 1)), x, oo) == oo


def test_sympyissue_15055():
    assert limit(n**3*((-n - 1)*sin(1/n) + (n + 2)*sin(1/(n + 1)))/(-n + 1), n, oo) == 1


def test_sympyissue_15146():
    assert limit((n/2)*(-2*n**3 - 2*(n**3 - 1)*n**2*digamma(n**3 + 1) +
                        2*(n**3 - 1)*n**2*digamma(n**3 + n + 1) +
                        n + 3), n, oo) == Rational(1, 3)


def test_sympyissue_15323():
    assert limit(((1 - 1/x)**x).diff(x), x, 1) == 1


def test_sympyissue_15984():
    assert limit((-x + log(exp(x) + 1))/x, x, oo, dir=1) == 0


def test_sympyissue_16222():
    assert limit(exp(x), x, 10000000) == exp(10000000)
    assert limit(exp(x), x, 100000000) == exp(100000000)
    assert limit(exp(x), x, 1000000000) == exp(1000000000)


@pytest.mark.timeout(20)
def test_sympyissue_15282():
    assert limit((x**2000 - (x + 1)**2000)/x**1999, x, oo) == -2000


def test_sympyissue_16722():
    n, z = symbols('n z')
    assert isinstance(limit(binomial(n + z, n)*n**-z, n, oo), Limit)

    z = symbols('z', positive=True)
    assert limit(binomial(n + z, n)*n**-z, n, oo) == 1/gamma(z + 1)

    n = symbols('n', positive=True, integer=True)
    z = symbols('z', positive=True)
    assert limit(binomial(n + z, n)*n**-z, n, oo) == 1/gamma(z + 1)

    n, z = symbols('n z', positive=True, integer=True)
    assert limit(binomial(n + z, n)*n**-z, n, oo) == 1/gamma(z + 1)


def test_sympyissue_15673():
    p = symbols('p')
    alpha = symbols('α', positive=True)

    e = Limit(4*pi*p**(-alpha)*(p**3 - p**alpha)/(alpha - 3), p, 0)
    assert isinstance(e.doit(), Limit)  # but see diofant/diofant#425


def test_sympyissue_17380():
    assert limit(x*(((x + 1)**2 + 1)/(x**2 + 1) - 1), x, oo) == 2


def test_sympyissue_17431():
    assert limit(((n + 1) + 1)/(((n + 1) + 2)*factorial(n + 1)) *
                 (n + 2)*factorial(n)/(n + 1), n, oo) == 0
    assert limit((n + 2)**2*factorial(n)/((n + 1)*(n + 3)*factorial(n + 1)),
                 n, oo) == 0

    # test from sympy/sympy#17434 (see also diofant/diofant#425):
    y = symbols('y', integer=True, positive=True)
    assert isinstance(limit(x*factorial(x)/factorial(x + y), x, oo), Limit)


def test_sympyissue_17792():
    n = Symbol('n', positive=True, integer=True)
    assert limit(factorial(n)/sqrt(n)*(E/n)**n, n, oo) == sqrt(2*pi)


def test_sympyissue_18118():
    assert limit(sign(x), x, 0) == +1
    assert limit(sign(x), x, 0, 1) == -1

    assert limit(sign(sin(x)), x, 0) == +1
    assert limit(sign(sin(x)), x, 0, 1) == -1


def test_sympyissue_6599():
    assert limit((x + cos(x))/x, x, oo) == 1


def test_sympyissue_18176():
    x = Symbol('x', real=True, positive=True)
    n = Symbol('n', integer=True, positive=True)
    e = x**n - x**(n - k)
    assert limit(e.subs({k: 0}), x, oo) == 0
    assert limit(e.subs({k: 1}), x, oo) == oo


def test_sympyissue_18306():
    assert limit(sin(sqrt(x))/sqrt(sin(x)), x, 0) == 1


def test_sympyissue_18378():
    assert limit(log(exp(3*x) + x)/log(exp(x) + x**100), x, oo) == 3


def test_sympyissue_18399():
    assert limit((1 - x/2)**(3*x), x, oo) == oo
    assert limit((-x)**x, x, oo) == oo


def test_sympyissue_18452():
    assert limit(abs(log(x))**x, x, 0) == 1
    assert limit(abs(log(x))**x, x, 0, 1) == 1


def test_sympyissue_18482():
    assert limit(sqrt(x**2 + 6*x) + (x**3 + x**2)/(x**2 + 1), x, -oo) == -2
    assert limit((x**3 + x**2 + sqrt(x*(x + 6))*(x**2 + 1))/(x**2 + 1),
                 x, -oo) == -2
    assert limit((2*exp(3*x)/(exp(2*x) + 1))**(1/x), x, oo) == E


def test_sympyissue_18501():
    assert limit(abs(log(x - 1)**3 - 1), x, 1) == oo


def test_sympyissue_18508():
    assert limit(sin(x)/sqrt(1 - cos(x)), x, 0) == sqrt(2)


def test_sympyissue_18707():
    p = Symbol('p', positive=True, real=True)

    assert limit(1/p**n, n, oo) == p**-oo


def test_sympyissue_18997():
    assert limit(abs(log(x)), x, 0) == oo
    assert limit(abs(log(abs(x))), x, 0) == oo


def test_sympyissue_18992():
    assert limit(x/(factorial(x)**(1/x)), x, oo) == E

    n = Symbol('n', positive=True, integer=True)

    assert limit(n/(factorial(n)**(1/n)), n, oo) == E


def test_sympyissue_19026():
    assert limit(abs(log(x) + 1)/log(x), x, oo) == 1


def test_sympyissue_19770():
    m = Symbol('m', extended_real=True)

    assert limit(cos(x*m)/x, x, oo) == 0


def test_sympyissue_19766():
    assert limit(2**(-x)*sqrt(4**(x + 1) + 1), x, oo) == 2


def test_sympyissue_14874():
    assert limit(besselk(0, x), x, oo) == 0


def test_sympyissue_20365():
    assert limit(((x + 1)**(1/x) - E)/x, x, 0) == -E/2


def test_sympyissue_20704():
    assert limit(x*(abs(1/x + y) - abs(y - 1/x))/2, x, 0) == 0


def test_sympyissue_21031():
    assert limit(((1 + x)**(1/x) -
                  (1 + 2*x)**(1/(2*x)))/asin(x), x, 0) == E/2


def test_sympyissue_21038():
    assert limit(sin(pi*x)/(3*x - 12), x, 4) == pi/3


def test_sympyissue_21029():
    e = (sinh(x) + cosh(x) - 1)/x/4
    ans0 = Rational(1, 4)
    ansz = (exp(z) - 1)/z/4

    assert limit(e, x, 0) == ans0
    assert limit(e, x, z) == ansz
    assert limit(ansz, z, 0) == ans0


def test_sympyissue_20578():
    e = abs(x)*sin(1/x)

    assert all(_ == 0 for _ in [limit(e, x, 0),
                                limit(e, x, 0, 1),
                                limit(e, x, 0, Reals)])


def test_sympyissue_19453():
    beta = Symbol('beta', real=True, positive=True)
    h = Symbol('h', real=True, positive=True)
    m = Symbol('m', real=True, positive=True)
    w = Symbol('omega', real=True, positive=True)
    g = Symbol('g', real=True, positive=True)

    q = 3*h**2*beta*g*exp(h*beta*w/2)
    p = m**2*w**2
    s = exp(h*beta*w) - 1
    z = (-q/(4*p*s) - q/(2*p*s**2) -
         q*(exp(h*beta*w) + 1)/(2*p*s**3) + exp(h*beta*w/2)/s)
    e = -diff(log(z), beta)

    assert limit(e - h*w/2, beta, oo) == 0
    assert limit(e.simplify() - h*w/2, beta, oo) == 0


def test_sympyissue_19442():
    pytest.raises(PoleError, lambda: limit(1/x, x, 0, Reals))


def test_sympyissue_21530():
    assert limit(sinh(n + 1)/sinh(n), n, oo) == E


def test_sympyissue_21550():
    r = (sqrt(5) - 1)/2
    assert limit((x - r)/(x**2 + x - 1), x, r).simplify() == sqrt(5)/5


def test_sympyissue_21606():
    assert limit(cos(x)/sign(x), x, pi) == -1
    assert limit(cos(x)/sign(x), x, pi, 1) == -1


def test_sympyissue_21756():
    e = (1 - exp(-2*I*pi*z))/(1 - exp(-2*I*pi*z/5))
    assert limit(e, z, 0) == 5


def test_sympyissue_21785():
    assert sqrt((-a**2 + x**2)/(1 - x**2)).limit(a, 1, 1) == I


def test_sympyissue_22220():
    e1 = sqrt(30)*atan(sqrt(30)*tan(x/2)/6)/30
    e2 = sqrt(30)*I*(-log(sqrt(2)*tan(x/2) - 2*sqrt(15)*I/5) +
                     +log(sqrt(2)*tan(x/2) + 2*sqrt(15)*I/5))/60

    assert limit(e1, x, -pi) == -sqrt(30)*pi/60
    assert limit(e2, x, -pi) == -sqrt(30)*pi/30

    assert limit(e1, x, -pi, 1) == sqrt(30)*pi/60
    assert limit(e2, x, -pi, 1) == 0


def test_sympyissue_22334():
    k = Symbol('k', positive=True)
    assert limit((n + 1)**k/((n + 1)**(k + 1) - n**(k + 1)), n, oo) == 1/(k + 1)


def test_sympyissue_22893():
    a, b = symbols('a b', positive=True)
    e = (a*exp(-a*x) + b*exp(-b*x))*exp(b*x)
    assert isinstance(limit(e, x, oo), Limit)


def test_sympyissue_22986():
    assert limit(acosh(1 + 1/x)*sqrt(x), x, oo) == sqrt(2)


def test_sympyissue_23231():
    assert limit((2**x - 2**-x)/(2**x + 2**-x), x, -oo) == -1


def test_issue_1164():
    # also sympy/sympy#14502
    assert limit(factorial(x) - x**x, x, oo) == -oo
    l = Limit(factorial(x) - x**oo, x, oo)
    assert l.doit() == l


def test_sympyissue_23266():
    assert limit(-0.295084971874737*exp(-18.9442719099992*x) +
                 5.29508497187474*exp(-1.05572809000084*x), x, oo) == 0


def test_sympyissue_7391():
    f = Function('f')
    func = x*y*y/(x*x + y**4)
    l = Limit(func.subs({y: f(x)}), x, 0)
    assert l.doit() == l
    assert l.subs({f: Lambda(x, sqrt(x))}).doit() == Rational(1, 2)


def test_issue_1216():
    assert (x/abs(sqrt(1 - x**2))).limit(x, oo) == 1
    assert ((4*x - 2)/abs(sqrt(4 - 4*(2*x - 1)**2))).limit(x, oo) == 1


def test_issue_1213():
    # also sympy/sympy#11496
    assert limit(erfc(log(1/x)), x, oo) == 2


def test_sympyissue_23319():
    assert limit(x*tan(pi/x), x, oo) == pi


def test_issue_1230():
    assert limit(log(x + sqrt(x**2 + 1)), x, I*oo) == oo


def test_sympyissue_8433():
    d = Symbol('d', positive=True)
    e = erf(1 - x/d)
    assert limit(e, x, oo) == -1
    assert limit(e.subs({d: 2}), x, oo) == -1


def test_sympyissue_13750():
    assert limit(erf(-x), x, oo) == -1
    assert limit(erf(1 - x), x, oo) == -1
    assert limit(erf(a - x), x, oo) == -1
    assert limit(erf(sqrt(x) - x), x, oo) == -1


def test_sympyissue_23836():
    p = Piecewise((x**3, x < 3), (-x**2, x > 3), (2, True))
    assert limit(p, x, 3) == -9
    assert limit(p, x, 3, 1) == 27


def test_sympyissue_24067():
    n = symbols('n', positive=True)
    e = (5*x)**(n + 2)/x**2
    assert limit(e, x, 0) == 0
    assert limit(e.simplify(), x, 0) == 0
    e = x**(n + 2)/x**2
    assert limit(e, x, 0) == 0


def test_sympyissue_24127():
    assert limit(Piecewise((sin(x), x < 0),
                           (Rational(1, 2), True)), x, 1) == Rational(1, 2)


def test_sympyissue_24210():
    assert limit(exp(x)/((1 + 1/x)**(x**2)), x, oo) == sqrt(E)


def test_sympyissue_24225():
    e = (log(1 + x) + log(1 + y)) / (x + y)
    assert limit(limit(e, x, 0, dir=Reals), y, 0, dir=Reals) == 1
    assert limit(limit(e, y, 0, dir=Reals), x, 0, dir=Reals) == 1


def test_sympyissue_24331():
    z = Symbol('z', complex=True)
    assert limit(log(z), z, 0) == -oo

    z = Symbol('z', infinite=True)
    assert limit(log(z), z, 0) == -oo


def test_sympyissue_24386():
    assert limit((x + y)/(exp(z) + 1) - y, z, -oo) == x



from ..core import Integer, Mul, nan
from ..functions import exp, log
from ..polys import quo, roots
from ..simplify.powsimp import powsimp
from .expr_with_intlimits import ExprWithIntLimits


class Product(ExprWithIntLimits):
    r"""
    Represents an unevaluated products.

    ``Product`` represents a finite or infinite product, with the first
    argument being the general form of terms in the series (which
    usually depend on the bound variable ``symbol``), and the second
    argument being ``(symbol, start, end)``, with ``symbol`` taking
    all integer values from ``start`` through ``end`` (inclusive).

    Notes
    =====

    We follow the the analogue of the summation convention described by
    Karr :cite:`Karr1981summation`, adopted by the
    :class:`~diofant.concrete.summations.Sum`:

    .. math::

        \prod\limits_{i=m}^n f_i = \frac{1}{\prod\limits_{i=n+1}^{m-1}f_i}

    Examples
    ========

    >>> Product(k**2, (k, 1, m)).doit()
    factorial(m)**2

    Products with the lower limit being larger than the upper one:

    >>> Product(1/k, (k, 6, 1)).doit()
    120
    >>> Product(k, (k, 2, 5)).doit()
    120

    The empty product:

    >>> Product(k, (k, n, n-1)).doit()
    1

    See Also
    ========

    diofant.concrete.summations.Sum
    diofant.concrete.summations.summation
    product

    References
    ==========

    * https://en.wikipedia.org/wiki/Multiplication#Capital_pi_notation
    * https://en.wikipedia.org/wiki/Empty_product

    """

    def _eval_rewrite_as_Sum(self, *args):
        from .summations import Sum
        return exp(Sum(log(self.function), *self.limits))

    @property
    def term(self):
        return self.args[0]
    function = term

    def _eval_is_zero(self):
        if self.term.is_zero:
            return True

    def doit(self, **hints):
        f = self.function
        for index, limit in enumerate(self.limits):
            i, a, b = limit
            dif = b - a
            if dif.is_Integer and dif < 0:
                a, b = b + 1, a - 1
                f = 1 / f

            g = self._eval_product(f, (i, a, b))
            if g in (None, nan):
                return self.func(powsimp(f), *self.limits[index:])
            else:
                f = g

        if hints.get('deep', True):
            return f.doit(**hints)
        else:
            return powsimp(f)

    def _eval_adjoint(self):
        if self.is_commutative:
            return self.func(self.function.adjoint(), *self.limits)

    def _eval_conjugate(self):
        return self.func(self.function.conjugate(), *self.limits)

    def _eval_product(self, term, limits):
        from ..functions import KroneckerDelta, RisingFactorial
        from .delta import _has_simple_delta, deltaproduct
        from .summations import summation

        (k, a, n) = limits

        if k not in term.free_symbols:
            if (term - 1).is_zero:
                return Integer(1)
            return term**(n - a + 1)

        if a == n:
            return term.subs({k: a})

        if term.has(KroneckerDelta) and _has_simple_delta(term, limits[0]):
            return deltaproduct(term, limits)

        dif = n - a
        if dif.is_Integer:
            return Mul(*[term.subs({k: a + i}) for i in range(dif + 1)])

        elif term.is_polynomial(k):
            poly = term.as_poly(k)

            A = B = Q = Integer(1)

            all_roots = roots(poly)

            M = 0
            for r, m in all_roots.items():
                M += m
                A *= RisingFactorial(a - r, n - a + 1)**m
                Q *= (k - r)**m

            if 0 < M < poly.degree():
                arg = quo(poly, Q.as_poly(k)).as_expr(k)
                B = self.func(arg, (k, a, n)).doit()

            if M > 0:
                return poly.LC()**(n - a + 1) * A * B

        elif term.is_Add:
            p, q = term.as_numer_denom()
            q = self._eval_product(q, (k, a, n))
            if q.is_Number:
                # There is expression, which couldn't change by
                # as_numer_denom(). E.g. n**(2/3) + 1 --> (n**(2/3) + 1, 1).
                p = sum(self._eval_product(i, (k, a, n)) for i in p.as_coeff_Add())
            else:
                p = self._eval_product(p, (k, a, n))
            return p / q

        elif term.is_Mul:
            exclude, include = [], []

            for t in term.args:
                p = self._eval_product(t, (k, a, n))

                if p is not None:
                    exclude.append(p)
                else:
                    include.append(t)

            if exclude:
                arg = term._new_rawargs(*include)
                A = Mul(*exclude)
                B = self.func(arg, (k, a, n)).doit()
                return A * B

        elif term.is_Pow:
            if not term.base.has(k):
                s = summation(term.exp, (k, a, n))

                return term.base**s
            elif not term.exp.has(k):
                p = self._eval_product(term.base, (k, a, n))

                if p is not None:
                    return p**term.exp

    def _eval_simplify(self, ratio, measure):
        from ..simplify.simplify import product_simplify
        return product_simplify(self)

    def _eval_transpose(self):
        if self.is_commutative:
            return self.func(self.function.transpose(), *self.limits)

    def reverse_order(self, *indices):
        r"""
        Reverse the order of a limit in a Product.

        Parameters
        ==========

        \*indices : list
            The selectors in the argument ``indices`` specify some indices whose
            limits get reversed.  These selectors are either variable names or
            numerical indices counted starting from the inner-most limit tuple.

        Examples
        ========

        >>> P = Product(x, (x, a, b))
        >>> Pr = P.reverse_order(x)
        >>> Pr
        Product(1/x, (x, b + 1, a - 1))
        >>> Pr = Pr.doit()
        >>> Pr
        1/RisingFactorial(b + 1, a - b - 1)
        >>> simplify(Pr)
        gamma(b + 1)/gamma(a)
        >>> P = P.doit()
        >>> P
        RisingFactorial(a, -a + b + 1)
        >>> simplify(P)
        gamma(b + 1)/gamma(a)

        While one should prefer variable names when specifying which limits
        to reverse, the index counting notation comes in handy in case there
        are several symbols with the same name.

        >>> s = Sum(x*y, (x, a, b), (y, c, d))
        >>> s
        Sum(x*y, (x, a, b), (y, c, d))
        >>> s0 = s.reverse_order(0)
        >>> s0
        Sum(-x*y, (x, b + 1, a - 1), (y, c, d))
        >>> s1 = s0.reverse_order(1)
        >>> s1
        Sum(x*y, (x, b + 1, a - 1), (y, d + 1, c - 1))

        Of course we can mix both notations:

        >>> Sum(x*y, (x, a, b), (y, 2, 5)).reverse_order(x, 1)
        Sum(x*y, (x, b + 1, a - 1), (y, 6, 1))
        >>> Sum(x*y, (x, a, b), (y, 2, 5)).reverse_order(y, x)
        Sum(x*y, (x, b + 1, a - 1), (y, 6, 1))

        See Also
        ========

        diofant.concrete.expr_with_intlimits.ExprWithIntLimits.index,
        diofant.concrete.expr_with_intlimits.ExprWithIntLimits.reorder_limit,
        diofant.concrete.expr_with_intlimits.ExprWithIntLimits.reorder

        References
        ==========

        * :cite:`Karr1981summation`

        """
        l_indices = list(indices)

        for i, indx in enumerate(l_indices):
            if not isinstance(indx, int):
                l_indices[i] = self.index(indx)

        e = 1
        limits = []
        for i, limit in enumerate(self.limits):
            l = limit
            if i in l_indices:
                e = -e
                l = (limit[0], limit[2] + 1, limit[1] - 1)
            limits.append(l)

        return Product(self.function ** e, *limits)


def product(*args, **kwargs):
    r"""
    Compute the product.

    The notation for symbols is similar to the notation used in Sum or
    Integral. product(f, (i, a, b)) computes the product of f with
    respect to i from a to b, i.e.,

    ::

                                     b
                                   _____
        product(f(n), (i, a, b)) = |   | f(n)
                                   |   |
                                   i = a

    If it cannot compute the product, it returns an unevaluated Product object.
    Repeated products can be computed by introducing additional symbols tuples::

    >>> i = symbols('i', integer=True)

    >>> product(i, (i, 1, k))
    factorial(k)
    >>> product(m, (i, 1, k))
    m**k
    >>> product(i, (i, 1, k), (k, 1, n))
    Product(factorial(k), (k, 1, n))

    """
    prod = Product(*args, **kwargs)

    if isinstance(prod, Product):
        return prod.doit(deep=False)
    else:
        return prod



"""Tests of tools for setting up interactive IPython sessions."""

import ast
import uuid

from diofant.interactive.session import (IntegerDivisionWrapper,
                                         unicode_identifiers)


__all__ = ()


def test_IntegerDivisionWrapper():
    tree = ast.parse('1/3')
    tree2 = ast.parse('Fraction(1, 3)')
    dump = ast.dump(tree2)
    tree_new = IntegerDivisionWrapper().visit(tree)
    assert ast.dump(tree_new) == dump

    tree = ast.parse('1 + 3')
    tree_new = IntegerDivisionWrapper().visit(tree)
    assert ast.dump(tree_new) == ast.dump(tree)

    tree = ast.parse('-1/3')
    tree2 = ast.parse('Fraction(-1, 3)')
    dump = ast.dump(tree2)
    tree_new = IntegerDivisionWrapper().visit(tree)
    assert ast.dump(tree_new) == dump

    tree = ast.parse('2**3/7')
    tree_new = IntegerDivisionWrapper().visit(tree)
    assert ast.dump(tree_new) == ast.dump(tree)

    tree = ast.parse('(3 + 5)/7')
    tree_new = IntegerDivisionWrapper().visit(tree)
    assert ast.dump(tree_new) == ast.dump(tree)

    tree = ast.parse('2**x/3')
    tree_new = IntegerDivisionWrapper().visit(tree)
    assert ast.dump(tree_new) == ast.dump(tree)


def test_unicode_identifiers(monkeypatch):
    class dummy_uuid4:
        _counter = 0

        @property
        def hex(self):
            self._counter += 1
            return str(self._counter)

    monkeypatch.setattr(uuid, 'uuid4', dummy_uuid4)

    assert unicode_identifiers(['ℕ = 1\n', 'N = 2\n',
                                'N + ℕ']) == ['_1 =1 \n', 'N =2 \n',
                                              'N +_1 ']



"""Tools for solving inequalities and systems of inequalities."""

import collections
import itertools

from ..core import Dummy, Eq, Ge, Gt, Integer, Le, Lt, Ne, S, Symbol, oo
from ..core.compatibility import iterable
from ..core.relational import Relational
from ..functions import Abs, Max, Min, Piecewise, sign
from ..logic import And, Or, false, true
from ..matrices import Matrix, diag
from ..polys import PolificationFailed, Poly, parallel_poly_from_expr
from ..polys.polyutils import _nsort
from ..sets import FiniteSet, Interval, Union
from ..utilities import filldedent, ordered


__all__ = 'reduce_inequalities',


def canonicalize_inequalities(eqs):
    """Canonicalize system of inequalities to have only Lt/Le."""
    eqs = set(eqs)

    # Canonicalize constraints, Ne -> pair Lt, Eq -> pair Le
    eqs |= {Lt(*e.args) for e in eqs if isinstance(e, Ne)}
    eqs |= {Lt(e.rhs, e.lhs) for e in eqs if isinstance(e, Ne)}
    eqs |= {Le(*e.args) for e in eqs if isinstance(e, Eq)}
    eqs |= {Le(e.rhs, e.lhs) for e in eqs if isinstance(e, Eq)}
    eqs -= {e for e in eqs if isinstance(e, (Ne, Eq))}

    # Gt/Ge -> Lt, Le
    eqs = {e.reversed if e.func in (Gt, Ge) else e for e in eqs}

    # Now we have only Lt/Le
    return list(ordered(e.func(e.lhs - e.rhs, 0) for e in eqs))


def fourier_motzkin(A, b, c, j):
    """
    Fourier-Motzkin elimination for `j`-th variable.

    Parameters
    ==========

    A : Matrix
        The coefficients of the system.
    b : Matrix
        The constant terms in the right hand side of relations.
    c : Matrix
        The vector of boolean elements, which determine the
        type of relation (1 for Le and 0 - for Lt).
    j : int
        The variable index.

    Example
    =======

    >>> A = Matrix([[-1, 0], [2, 4], [1, -2]])
    >>> b = Matrix([-1, 14, -1])
    >>> c = Matrix([1, 1, 1])
    >>> fourier_motzkin(A, b, c, 0)
    (Matrix([
    [0,  4],
    [0, -2]]), Matrix([
    [12],
    [-2]]), Matrix([
    [1],
    [1]]))

    References
    ==========

    * :cite:`Schrijver1998theory`, pp. 155–156.

    """
    m = A.rows
    rows = [[], [], []]
    D, d, k = [Matrix()]*3

    assert m == b.rows == c.rows
    assert all(_.is_comparable for _ in A)

    for i, a in enumerate(A[:, j]):
        rows[int(sign(a) + 1)].append(i)

    for p in itertools.chain(rows[1], itertools.product(*rows[::2])):
        if p in rows[1]:
            D = D.col_join(A[p, :])
            d = d.col_join(Matrix([b[p]]))
            k = k.col_join(Matrix([c[p]]))
        else:
            s, t = p
            D = D.col_join(A[t, j]*A[s, :] - A[s, j]*A[t, :])
            d = d.col_join(Matrix([A[t, j]*b[s] - A[s, j]*b[t]]))
            k = k.col_join(Matrix([c[s] and c[t]]))

    return D, d, k


def solve_linear_inequalities(eqs, *gens, **args):
    """
    Solve system of linear inequalities.

    Examples
    ========

    >>> solve_linear_inequalities([x >= 0, 2*x + 4*y <= 14, x - 2*y <= 1])
    (x >= 0) & (x <= 4) & (y >= x/2 - 1/2) & (y <= -x/2 + 7/2)

    """
    assert all(e.is_Relational for e in eqs)

    eqs = canonicalize_inequalities(eqs)

    polys, opt = parallel_poly_from_expr([e.lhs for e in eqs], *gens, **args)

    if not all(p.is_linear for p in polys):
        raise ValueError(f'Got non-linear inequality in {eqs}')

    gens = Matrix(opt.gens)
    A = Matrix([[p.coeff_monomial(x) for x in gens] for p in polys])
    b = Matrix([-p.coeff_monomial(1) for p in polys])
    c = Matrix([e.func is Le for e in eqs])
    res = []
    failed = []
    op_map = {(1, 1): Le, (1, 0): Lt, (0, 1): Ge, (0, 0): Gt}

    for i, g in reversed(list(enumerate(gens))):
        D, d, e = fourier_motzkin(A, b, c, i)

        if not D:
            failed.append(i)
            continue

        gens_g = gens.copy()
        gens_g[i] = 0

        for j, (r, x) in enumerate(zip(b - A*gens_g, c)):
            gc = A[j, i]
            if not gc:
                continue

            res.append(op_map[(int(gc > 0), int(x))](g, r/gc))

        A, b, c = D, d, e

    if not A.is_zero:
        i = failed.pop(0)
        g = gens[i]
        gens_g = gens.copy()
        gens_g[i] = 0
        strict = []
        non_strict = []

        for r, x in zip(diag(*A[:, i])**-1*(b - A*gens_g), c):
            if x:
                non_strict.append(r)
            else:
                strict.append(r)

        pos = int(A[0, i] > 0)
        other_op = Min if pos else Max

        if strict and non_strict:
            a, b = other_op(*non_strict), other_op(*strict)
            opn, ops = op_map[(pos, 1)], op_map[(pos, 0)]
            res.append(Or(And(opn(g, a), ops(a, b)), And(ops(g, b), opn(b, a))))
        else:
            both = non_strict + strict
            res.append(op_map[(pos, int(not strict))](g, other_op(*(both))))
    elif any(_ < 0 for _ in b):
        return false

    return And(*res)


def solve_poly_inequality(poly, rel):
    """
    Solve a polynomial inequality with rational coefficients.

    Examples
    ========

    >>> solve_poly_inequality(x.as_poly(), '==')
    [{0}]
    >>> solve_poly_inequality((x**2 - 1).as_poly(), '!=')
    [[-oo, -1), (-1, 1), (1, oo]]
    >>> solve_poly_inequality((x**2 - 1).as_poly(), '==')
    [{-1}, {1}]

    See Also
    ========

    solve_poly_inequalities

    """
    if not isinstance(poly, Poly):
        raise ValueError('`poly` should be a Poly instance')
    if rel not in {'>', '<', '>=', '<=', '==', '!='}:
        raise ValueError(f'Invalid relational operator symbol: {rel!r}')
    if poly.is_number:
        t = Relational(poly.as_expr(), 0, rel)
        if t == true:
            return [S.ExtendedReals]
        elif t == false:
            return [S.EmptySet]
        else:
            raise NotImplementedError(f"Couldn't determine truth value of {t}")

    reals, intervals = poly.real_roots(multiple=False), []

    if rel == '==':
        for root, _ in reals:
            interval = Interval(root, root)
            intervals.append(interval)
    elif rel == '!=':
        left = -oo

        for right, _ in reals + [(oo, 1)]:
            interval = Interval(left, right, left.is_finite, right.is_finite)
            intervals.append(interval)
            left = right
    else:
        sign = +1 if poly.LC() > 0 else -1
        eq_sign, equal = None, False

        if rel == '>':
            eq_sign = +1
        elif rel == '<':
            eq_sign = -1
        elif rel == '>=':
            eq_sign, equal = +1, True
        else:
            eq_sign, equal = -1, True

        right, right_open = oo, False

        for left, multiplicity in reversed(reals):
            if multiplicity % 2:
                if sign == eq_sign:
                    intervals.insert(0, Interval(left, right, not equal and left.is_finite, right_open and right.is_finite))

                sign, right, right_open = -sign, left, not equal
            else:
                if sign == eq_sign and not equal:
                    intervals.insert(0, Interval(left, right, left.is_finite, right_open and right.is_finite))
                    right, right_open = left, True
                elif sign != eq_sign and equal:
                    intervals.insert(0, Interval(left, left))

        if sign == eq_sign:
            intervals.insert(0, Interval(-oo, right, False, right_open and right.is_finite))

    return intervals


def solve_poly_inequalities(polys):
    """
    Solve polynomial inequalities with rational coefficients.

    Examples
    ========

    >>> solve_poly_inequalities((((+x**2 - 3).as_poly(), '>'),
    ...                          ((-x**2 + 1).as_poly(), '>')))
    [-oo, -sqrt(3)) U (-1, 1) U (sqrt(3), oo]

    """
    return Union(*[solve_poly_inequality(*p) for p in polys])


def solve_rational_inequalities(eqs):
    """
    Solve a system of rational inequalities with rational coefficients.

    Examples
    ========

    >>> solve_rational_inequalities([[(((-x + 1).as_poly(),
    ...                                 Integer(1).as_poly(x)), '>='),
    ...                               (((-x + 1).as_poly(),
    ...                                 Integer(1).as_poly(x)), '<=')]])
    {1}

    >>> solve_rational_inequalities([[((x.as_poly(), Integer(1).as_poly(x)), '!='),
    ...                               (((-x + 1).as_poly(),
    ...                                 Integer(1).as_poly(x)), '>=')]])
    [-oo, 0) U (0, 1]

    See Also
    ========

    solve_poly_inequality

    """
    result = S.EmptySet

    for eq in eqs:
        global_intervals = [S.ExtendedReals]

        for (numer, denom), rel in eq:
            intervals = []

            for numer_interval in solve_poly_inequality(numer*denom, rel):
                for global_interval in global_intervals:
                    interval = numer_interval & global_interval

                    if interval is not S.EmptySet:
                        intervals.append(interval)

            global_intervals = intervals

            intervals = []

            for global_interval in global_intervals:
                for denom_interval in solve_poly_inequality(denom, '=='):
                    global_interval -= denom_interval

                if global_interval is not S.EmptySet:
                    intervals.append(global_interval)

            global_intervals = intervals

            if not global_intervals:
                break

            intervals = []

            for interval in global_intervals:
                intervals.append(interval)

            global_intervals = intervals

        for interval in global_intervals:
            result |= interval

    return result


def reduce_rational_inequalities(exprs, gen, relational=True):
    """
    Reduce a system of rational inequalities with rational coefficients.

    Examples
    ========

    >>> reduce_rational_inequalities([[x**2 <= 0]], x)
    Eq(x, 0)
    >>> reduce_rational_inequalities([[x + 2 > 0]], x)
    -2 < x
    >>> reduce_rational_inequalities([[(x + 2, '>')]], x)
    -2 < x
    >>> reduce_rational_inequalities([[x + 2]], x)
    Eq(x, -2)

    """
    exact = True
    eqs = []
    solution = S.ExtendedReals if exprs else S.EmptySet
    for _exprs in exprs:
        _eqs = []

        for expr in _exprs:
            if isinstance(expr, tuple):
                expr, rel = expr
            else:
                if expr.is_Relational:
                    expr, rel = expr.lhs - expr.rhs, expr.rel_op
                else:
                    rel = '=='

            if expr == true:
                numer, denom, rel = Integer(0), Integer(1), '=='
            elif expr == false:
                numer, denom, rel = Integer(1), Integer(1), '=='
            else:
                numer, denom = expr.together().as_numer_denom()

            (numer, denom), opt = parallel_poly_from_expr((numer, denom), gen)

            if not opt.domain.is_Exact:
                numer, denom, exact = numer.to_exact(), denom.to_exact(), False

            domain = opt.domain.get_exact()

            if not (domain.is_IntegerRing or domain.is_RationalField):
                expr = numer/denom
                expr = Relational(expr, 0, rel)
                solution &= solve_univariate_inequality(expr, gen, relational=False)
            else:
                _eqs.append(((numer, denom), rel))

        if _eqs:
            eqs.append(_eqs)

    if eqs:
        solution &= solve_rational_inequalities(eqs)

    if not exact:
        solution = solution.evalf()

    if relational:
        solution = solution.as_relational(gen)

    return solution


def reduce_piecewise_inequality(expr, rel, gen):
    """
    Reduce an inequality with nested piecewise functions.

    Examples
    ========

    >>> reduce_piecewise_inequality(abs(x - 5) - 3, '<', x)
    (2 < x) & (x < 8)
    >>> reduce_piecewise_inequality(abs(x + 2)*3 - 13, '<', x)
    (-19/3 < x) & (x < 7/3)

    >>> reduce_piecewise_inequality(Piecewise((1, x < 1),
    ...                                       (3, True)) - 1, '>', x)
    1 <= x

    See Also
    ========

    reduce_piecewise_inequalities

    """
    if gen.is_extended_real is False:
        raise TypeError(filldedent("""
            can't solve inequalities with piecewise
            functions containing non-real variables"""))

    def _bottom_up_scan(expr):
        exprs = []

        if expr.is_Add or expr.is_Mul:
            op = expr.func

            for arg in expr.args:
                _exprs = _bottom_up_scan(arg)

                if not exprs:
                    exprs = _exprs
                else:
                    args = []

                    for expr, conds in exprs:
                        for _expr, _conds in _exprs:
                            args.append((op(expr, _expr), conds + _conds))

                    exprs = args
        elif expr.is_Pow:
            n = expr.exp

            if not n.is_Integer:
                raise NotImplementedError('only integer powers are supported')

            _exprs = _bottom_up_scan(expr.base)

            for expr, conds in _exprs:
                exprs.append((expr**n, conds))
        elif isinstance(expr, Abs):
            _exprs = _bottom_up_scan(expr.args[0])

            for expr, conds in _exprs:
                exprs.append((+expr, conds + [Ge(expr, 0)]))
                exprs.append((-expr, conds + [Lt(expr, 0)]))
        elif isinstance(expr, Piecewise):
            for a in expr.args:
                _exprs = _bottom_up_scan(a.expr)

                for ex, conds in _exprs:
                    if a.cond != true:
                        exprs.append((ex, conds + [a.cond]))
                    else:
                        oconds = [c[1] for c in expr.args if c[1] != true]
                        exprs.append((ex, conds + [And(*[~c for c in oconds])]))
        else:
            exprs = [(expr, [])]

        return exprs

    exprs = _bottom_up_scan(expr)

    mapping = {'<': '>', '<=': '>='}
    inequalities = []

    for expr, conds in exprs:
        if rel not in mapping:
            expr = Relational(+expr, 0, rel)
        else:
            expr = Relational(-expr, 0, mapping[rel])

        inequalities.append([expr] + conds)

    return reduce_rational_inequalities(inequalities, gen)


def reduce_piecewise_inequalities(exprs, gen):
    """
    Reduce a system of inequalities with nested piecewise functions.

    Examples
    ========

    >>> reduce_piecewise_inequalities([(abs(3*x - 5) - 7, '<'),
    ...                                (abs(x + 25) - 13, '>')], x)
    (-2/3 < x) & (x < 4) & ((-12 < x) | (x < -38))
    >>> reduce_piecewise_inequalities([(abs(x - 4) + abs(3*x - 5) - 7, '<')], x)
    (1/2 < x) & (x < 4)

    See Also
    ========

    reduce_piecewise_inequality

    """
    return And(*[reduce_piecewise_inequality(expr, rel, gen)
                 for expr, rel in exprs])


def solve_univariate_inequality(expr, gen, relational=True):
    """
    Solves a real univariate inequality.

    Examples
    ========

    >>> solve_univariate_inequality(x**2 >= 4, x)
    (2 <= x) | (x <= -2)
    >>> solve_univariate_inequality(x**2 >= 4, x, relational=False)
    [-oo, -2] U [2, oo]

    """
    from ..simplify import simplify
    from .solvers import denoms, solve

    e = expr.lhs - expr.rhs
    solns = solve(e, gen)
    singularities = []
    for d in denoms(e):
        singularities.extend(solve(d, gen))
    solns = [s[gen] for s in solns]
    singularities = [s[gen] for s in singularities]

    include_x = expr.func(0, 0)

    def valid(x):
        v = e.subs({gen: x})
        try:
            r = expr.func(v, 0)
        except TypeError:
            r = false
        r = simplify(r)
        if r in (true, false):
            return r
        elif v.is_comparable is False:
            return False
        else:
            raise NotImplementedError

    start = -oo
    sol_sets = [S.EmptySet]
    reals = _nsort(set(solns + singularities), separated=True)[0]
    for x in reals:
        end = x

        if end in [-oo, oo]:
            if valid(Integer(0)):
                sol_sets.append(Interval(start, oo, start in reals, end == oo))
                break

        if valid((start + end)/2 if start != -oo else end - 1):
            sol_sets.append(Interval(start, end, start.is_finite is not False, end.is_finite is not False))

        if x in singularities:
            singularities.remove(x)
        elif include_x:
            sol_sets.append(FiniteSet(x))

        start = end

    end = oo

    if valid(start + 1):
        sol_sets.append(Interval(start, end, True, end in reals))

    rv = Union(*sol_sets)

    rel_map = {Lt: Le, Gt: Ge}

    for t in [oo, -oo]:
        try:
            rel = rel_map.get(expr.func, expr.func)
            if rv.contains(t) is true and rel(e.limit(gen, t)) is false:
                rv -= FiniteSet(t)
        except TypeError:
            pass

    return rv if not relational else rv.as_relational(gen)


def _reduce_inequalities(inequalities, symbols):
    if len(symbols) > 1:
        try:
            return solve_linear_inequalities(inequalities, *symbols)
        except (PolificationFailed, ValueError):
            pass

    rat_part = collections.defaultdict(list)
    pw_part = rat_part.copy()
    other = []

    for inequality in inequalities:
        if inequality == true:
            continue
        if inequality == false:
            return false

        expr, rel = inequality.lhs, inequality.rel_op  # rhs is 0

        # check for gens using atoms which is more strict than free_symbols to
        # guard against EX domain which won't be handled by
        # reduce_rational_inequalities
        gens = expr.atoms(Dummy, Symbol)

        if len(gens) == 1:
            gen = gens.pop()
        else:
            common = expr.free_symbols & set(symbols)
            if len(common) == 1:
                gen = common.pop()
                other.append(solve_univariate_inequality(Relational(expr, 0, rel), gen))
                continue
            raise NotImplementedError('Solving multivariate inequalities is '
                                      'implemented only for linear case yet.')

        if expr.is_rational_function(gen):
            rat_part[gen].append((expr, rel))
        else:
            components = set(expr.find(lambda u: u.has(gen) and
                                       (u.is_Function or u.is_Pow and
                                        not u.exp.is_Integer)))
            if components and all(isinstance(i, (Abs, Piecewise)) for i in components):
                pw_part[gen].append((expr, rel))
            else:
                other.append(solve_univariate_inequality(Relational(expr, 0, rel), gen))

    rat_reduced = []
    pw_reduced = []

    for gen, exprs in rat_part.items():
        rat_reduced.append(reduce_rational_inequalities([exprs], gen))

    for gen, exprs in pw_part.items():
        pw_reduced.append(reduce_piecewise_inequalities(exprs, gen))

    return And(*(rat_reduced + pw_reduced + other))


def reduce_inequalities(inequalities, symbols=[]):
    """
    Reduces a system of inequalities or equations.

    Examples
    ========

    >>> reduce_inequalities(0 <= x + 3, [])
    -3 <= x
    >>> reduce_inequalities(0 <= x + y*2 - 1, [x])
    -2*y + 1 <= x

    See Also
    ========

    diofant.solvers.solvers.solve : solve algebraic equations

    """
    if not iterable(inequalities):
        inequalities = [inequalities]

    # prefilter
    keep = []
    for i in inequalities:
        if isinstance(i, Relational):
            i = i.func(i.lhs.as_expr() - i.rhs.as_expr(), 0)
        elif i not in (True, False):
            i = Eq(i, 0)
        if i == true:
            continue
        if i == false:
            return false
        keep.append(i)
    inequalities = keep
    del keep

    gens = set().union(*[i.free_symbols for i in inequalities])

    if not iterable(symbols):
        symbols = [symbols]
    symbols = ordered(set(symbols) or gens)

    # make vanilla symbol real
    recast = {i: Dummy(i.name, extended_real=True)
              for i in gens if i.is_extended_real is None}
    inequalities = [i.xreplace(recast) for i in inequalities]
    symbols = ordered(i.xreplace(recast) for i in symbols)

    # solve system
    rv = _reduce_inequalities(inequalities, list(symbols))

    # restore original symbols and return
    return rv.xreplace({v: k for k, v in recast.items()})



import itertools
import math

import mpmath
import pytest

import diofant
from diofant import (ITE, And, Float, Function, I, Integral, Lambda, Matrix,
                     Max, Min, Not, Or, Piecewise, Rational, cos, exp, false,
                     lambdify, oo, pi, sin, sqrt, symbols, true)
from diofant.abc import t, w, x, y, z
from diofant.external import import_module
from diofant.printing.lambdarepr import LambdaPrinter
from diofant.utilities.decorator import conserve_mpmath_dps
from diofant.utilities.lambdify import (MATH_TRANSLATIONS, MPMATH_TRANSLATIONS,
                                        NUMPY_TRANSLATIONS, _get_namespace,
                                        implemented_function, lambdastr)


__all__ = ()

MutableDenseMatrix = Matrix

numpy = import_module('numpy')
with_numpy = pytest.mark.skipif(numpy is None,
                                reason="Couldn't import numpy.")

# ================= Test different arguments =======================


def test_no_args():
    f = lambdify([], 1)
    pytest.raises(TypeError, lambda: f(-1))
    assert f() == 1


def test_single_arg():
    f = lambdify(x, 2*x)
    assert f(1) == 2


def test_list_args():
    f = lambdify([x, y], x + y)
    assert f(1, 2) == 3


def test_nested_args():
    # issue sympy/sympy#2790
    assert lambdify((x, (y, z)), x + y)(1, (2, 4)) == 3
    assert lambdify((x, (y, (w, z))), w + x + y + z)(1, (2, (3, 4))) == 10
    assert lambdify(x, x + 1, dummify=False)(1) == 2


def test_str_args():
    f = lambdify('x,y,z', 'z,y,x')
    assert f(3, 2, 1) == (1, 2, 3)
    assert f(1.0, 2.0, 3.0) == (3.0, 2.0, 1.0)
    # make sure correct number of args required
    pytest.raises(TypeError, lambda: f(0))


def test_own_namespace():
    def myfunc(x):
        return 1
    f = lambdify(x, sin(x), {'sin': myfunc})
    assert f(0.1) == 1
    assert f(100) == 1


def test_own_module():
    f = lambdify(x, sin(x), math)
    assert f(0) == 0.0


def test_bad_args():
    # no vargs given
    pytest.raises(TypeError, lambda: lambdify(1))
    # same with vector exprs
    pytest.raises(TypeError, lambda: lambdify([1, 2]))
    # reserved name
    pytest.raises(ValueError, lambda: lambdify((('__flatten_args__',),), 1))

    pytest.raises(NameError, lambda: lambdify(x, 1, 'spam'))


def test__get_namespace():
    pytest.raises(TypeError, lambda: _get_namespace(1))


def test_lambdastr():
    assert lambdastr(x, x**2) == 'lambda x: (x**2)'
    assert lambdastr(x, None, dummify=True).find('None') > 0


def test_atoms():
    # Non-Symbol atoms should not be pulled out from the expression namespace
    f = lambdify(x, pi + x, {'pi': 3.14})
    assert f(0) == 3.14
    f = lambdify(x, I + x, {'I': 1j})
    assert f(1) == 1 + 1j

# ================= Test different modules =========================

# high precision output of sin(0.2*pi) is used to detect if precision is lost unwanted


@conserve_mpmath_dps
def test_diofant_lambda():
    mpmath.mp.dps = 50
    sin02 = mpmath.mpf('0.19866933079506121545941262711838975037020672954020')
    f = lambdify(x, sin(x), 'diofant')
    assert f(x) == sin(x)
    prec = 1e-15
    assert -prec < f(Rational(1, 5)).evalf() - Float(str(sin02)) < prec


@conserve_mpmath_dps
def test_math_lambda():
    mpmath.mp.dps = 50
    sin02 = mpmath.mpf('0.19866933079506121545941262711838975037020672954020')
    f = lambdify(x, sin(x), 'math')
    prec = 1e-15
    assert -prec < f(0.2) - sin02 < prec

    # if this succeeds, it can't be a python math function
    pytest.raises(TypeError, lambda: f(x))


@conserve_mpmath_dps
def test_mpmath_lambda():
    mpmath.mp.dps = 50
    sin02 = mpmath.mpf('0.19866933079506121545941262711838975037020672954020')
    f = lambdify(x, sin(x), 'mpmath')
    prec = 1e-49  # mpmath precision is around 50 decimal places
    assert -prec < f(mpmath.mpf('0.2')) - sin02 < prec

    # if this succeeds, it can't be a mpmath function
    pytest.raises(TypeError, lambda: f(x))


@conserve_mpmath_dps
def test_number_precision():
    mpmath.mp.dps = 50
    sin02 = mpmath.mpf('0.19866933079506121545941262711838975037020672954020')
    f = lambdify(x, sin02, 'mpmath')
    prec = 1e-49  # mpmath precision is around 50 decimal places
    assert -prec < f(0) - sin02 < prec


@conserve_mpmath_dps
def test_mpmath_precision():
    mpmath.mp.dps = 100
    assert str(lambdify((), pi.evalf(100), 'mpmath')()) == str(pi.evalf(100))


# ================= Test Translations ==============================
# We can only check if all translated functions are valid. It has to be checked
# by hand if they are complete.


def test_math_transl():
    for sym, mat in MATH_TRANSLATIONS.items():
        assert sym in diofant.__dict__
        assert mat in math.__dict__


def test_mpmath_transl():
    for sym, mat in MPMATH_TRANSLATIONS.items():
        assert sym in diofant.__dict__ or sym == 'Matrix'
        assert mat in mpmath.__dict__


@with_numpy
def test_numpy_transl():
    for sym, nump in NUMPY_TRANSLATIONS.items():
        assert sym in diofant.__dict__
        assert nump in numpy.__dict__


@with_numpy
def test_numpy_translation_abs():
    f = lambdify(x, abs(x), 'numpy')
    assert f(-1) == 1
    assert f(1) == 1


# ================= Test some functions ============================


def test_exponentiation():
    f = lambdify(x, x**2)
    assert f(-1) == 1
    assert f(0) == 0
    assert f(1) == 1
    assert f(-2) == 4
    assert f(2) == 4
    assert f(2.5) == 6.25


def test_sqrt():
    f = lambdify(x, sqrt(x))
    assert f(0) == 0.0
    assert f(1) == 1.0
    assert f(4) == 2.0
    assert abs(f(2) - 1.414) < 0.001
    assert f(6.25) == 2.5


def test_trig():
    f = lambdify([x], [cos(x), sin(x)], 'math')
    d = f(pi)
    prec = 1e-11
    assert -prec < d[0] + 1 < prec
    assert -prec < d[1] < prec
    d = f(3.14159)
    prec = 1e-5
    assert -prec < d[0] + 1 < prec
    assert -prec < d[1] < prec

# ================= Test vectors ===================================


def test_vector_simple():
    f = lambdify((x, y, z), (z, y, x))
    assert f(3, 2, 1) == (1, 2, 3)
    assert f(1.0, 2.0, 3.0) == (3.0, 2.0, 1.0)
    # make sure correct number of args required
    pytest.raises(TypeError, lambda: f(0))


def test_vector_discontinuous():
    f = lambdify(x, (-1/x, 1/x))
    pytest.raises(ZeroDivisionError, lambda: f(0))
    assert f(1) == (-1.0, 1.0)
    assert f(2) == (-0.5, 0.5)
    assert f(-2) == (0.5, -0.5)


def test_trig_symbolic():
    f = lambdify([x], [cos(x), sin(x)], 'math')
    d = f(pi)
    assert abs(d[0] + 1) < 0.0001
    assert abs(d[1] - 0) < 0.0001


def test_trig_float():
    f = lambdify([x], [cos(x), sin(x)])
    d = f(3.14159)
    assert abs(d[0] + 1) < 0.0001
    assert abs(d[1] - 0) < 0.0001


def test_docs():
    f = lambdify(x, x**2)
    assert f(2) == 4
    f = lambdify([x, y, z], [z, y, x])
    assert f(1, 2, 3) == [3, 2, 1]
    f = lambdify(x, sqrt(x))
    assert f(4) == 2.0
    f = lambdify((x, y), sin(x*y)**2)
    assert f(0, 5) == 0


def test_math():
    f = lambdify((x, y), sin(x), modules='math')
    assert f(0, 5) == 0


def test_sin():
    f = lambdify(x, sin(x)**2)
    assert isinstance(f(2), float)
    f = lambdify(x, sin(x)**2, modules='math')
    assert isinstance(f(2), float)


def test_matrix():
    A = Matrix([[x, x*y], [sin(z) + 4, x**z]])
    sol = Matrix([[1, 2], [sin(3) + 4, 1]])
    f = lambdify((x, y, z), A, modules='diofant')
    assert f(1, 2, 3) == sol
    f = lambdify((x, y, z), (A, [A]), modules='diofant')
    assert f(1, 2, 3) == (sol, [sol])
    J = Matrix((x, x + y)).jacobian((x, y))
    v = Matrix((x, y))
    sol = Matrix([[1, 0], [1, 1]])
    assert lambdify(v, J, modules='diofant')(1, 2) == sol
    assert lambdify(v.T, J, modules='diofant')(1, 2) == sol


@with_numpy
def test_numpy_matrix():
    A = Matrix([[x, x*y], [sin(z) + 4, x**z]])
    sol_arr = numpy.array([[1, 2], [numpy.sin(3) + 4, 1]])
    # Lambdify array first, to ensure return to array as default
    f = lambdify((x, y, z), A, ['numpy'])
    numpy.testing.assert_allclose(f(1, 2, 3), sol_arr)
    # Check that the types are arrays and matrices
    assert isinstance(f(1, 2, 3), numpy.ndarray)


@with_numpy
def test_numpy_transpose():
    A = Matrix([[1, x], [0, 1]])
    f = lambdify(x, A.T, modules='numpy')
    numpy.testing.assert_array_equal(f(2), numpy.array([[1, 0], [2, 1]]))


@with_numpy
def test_numpy_inverse():
    A = Matrix([[1, x], [0, 1]])
    f = lambdify(x, A**-1, modules='numpy')
    numpy.testing.assert_array_equal(f(2), numpy.array([[1, -2], [0,  1]]))


@with_numpy
def test_numpy_old_matrix():
    A = Matrix([[x, x*y], [sin(z) + 4, x**z]])
    sol_arr = numpy.array([[1, 2], [numpy.sin(3) + 4, 1]])
    f = lambdify((x, y, z), A, [{'ImmutableMatrix': numpy.array}, 'numpy'])
    numpy.testing.assert_allclose(f(1, 2, 3), sol_arr)
    assert isinstance(f(1, 2, 3), numpy.ndarray)


@with_numpy
@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_sympyissue_11306():
    p = Piecewise((1/x, y < -1), (x, y <= 1), (1/x, True))
    lambdify([x, y], p, modules='numpy')(0, 1)


@with_numpy
def test_numpy_piecewise():
    pieces = Piecewise((x, x < 3), (x**2, x > 5), (0, True))
    f = lambdify(x, pieces, modules='numpy')
    numpy.testing.assert_array_equal(f(numpy.arange(10)),
                                     numpy.array([0, 1, 2, 0, 0, 0, 36, 49, 64, 81]))
    # If we evaluate somewhere all conditions are False, we should get back NaN
    nodef_func = lambdify(x, Piecewise((x, x > 0), (-x, x < 0)))
    numpy.testing.assert_array_equal(nodef_func(numpy.array([-1, 0, 1])),
                                     numpy.array([1, numpy.nan, 1]))


@with_numpy
def test_numpy_logical_ops():
    and_func = lambdify((x, y), And(x, y), modules='numpy')
    or_func = lambdify((x, y), Or(x, y), modules='numpy')
    not_func = lambdify(x, Not(x), modules='numpy')
    arr1 = numpy.array([True, True])
    arr2 = numpy.array([False, True])
    numpy.testing.assert_array_equal(and_func(arr1, arr2), numpy.array([False, True]))
    numpy.testing.assert_array_equal(or_func(arr1, arr2), numpy.array([True, True]))
    numpy.testing.assert_array_equal(not_func(arr2), numpy.array([True, False]))


@with_numpy
def test_numpy_matmul():
    xmat = Matrix([[x, y], [z, 1+z]])
    ymat = Matrix([[x**2], [abs(x)]])
    mat_func = lambdify((x, y, z), xmat*ymat, modules='numpy')
    numpy.testing.assert_array_equal(mat_func(0.5, 3, 4), numpy.array([[1.625], [3.5]]))
    numpy.testing.assert_array_equal(mat_func(-0.5, 3, 4), numpy.array([[1.375], [3.5]]))
    # Multiple matrices chained together in multiplication
    f = lambdify((x, y, z), xmat*xmat*xmat, modules='numpy')
    numpy.testing.assert_array_equal(f(0.5, 3, 4), numpy.array([[72.125, 119.25],
                                                                [159, 251]]))


def test_integral():
    f = Lambda(x, exp(-x**2))
    l = lambdify(x, Integral(f(x), (x, -oo, oo)), modules='diofant')
    assert l(x) == Integral(exp(-x**2), (x, -oo, oo))

# ================= Test symbolic ==================================


def test_sym_single_arg():
    f = lambdify(x, x * y)
    assert f(z) == z * y


def test_sym_list_args():
    f = lambdify([x, y], x + y + z)
    assert f(1, 2) == 3 + z


def test_sym_integral():
    f = Lambda(x, exp(-x**2))
    l = lambdify(x, Integral(f(x), (x, -oo, oo)), modules='diofant')
    assert l(y).doit() == sqrt(pi)


def test_namespace_order():
    # lambdify had a bug, such that module dictionaries or cached module
    # dictionaries would pull earlier namespaces into themselves.
    # Because the module dictionaries form the namespace of the
    # generated lambda, this meant that the behavior of a previously
    # generated lambda function could change as a result of later calls
    # to lambdify.
    n1 = {'f': lambda x: 'first f'}
    n2 = {'f': lambda x: 'second f',
          'g': lambda x: 'function g'}
    f = diofant.Function('f')
    g = diofant.Function('g')
    if1 = lambdify(x, f(x), modules=(n1, 'diofant'))
    assert if1(1) == 'first f'
    if2 = lambdify(x, g(x), modules=(n2, 'diofant'))
    assert if2(1) == 'function g'
    # previously gave 'second f'
    assert if1(1) == 'first f'


def test_imps():
    # Here we check if the default returned functions are anonymous - in
    # the sense that we can have more than one function with the same name
    f = implemented_function('f', lambda x: 2*x)
    g = implemented_function('f', math.sqrt)
    l1 = lambdify(x, f(x))
    l2 = lambdify(x, g(x))
    assert str(f(x)) == str(g(x))
    assert l1(3) == 6
    assert l2(3) == math.sqrt(3)
    # check that we can pass in a Function as input
    func = diofant.Function('myfunc')
    assert not hasattr(func, '_imp_')
    my_f = implemented_function(func, lambda x: 2*x)
    assert hasattr(func, '_imp_')
    assert hasattr(my_f, '_imp_')
    # Error for functions with same name and different implementation
    f2 = implemented_function('f', lambda x: x + 101)
    pytest.raises(ValueError, lambda: lambdify(x, f(f2(x))))


def test_imps_errors():
    # Test errors that implemented functions can return, and still be
    # able to form expressions.  See issue sympy/sympy#10810.
    for val, error_class in itertools.product((0, 0., 2, 2.0),
                                              (AttributeError, TypeError,
                                               ValueError)):

        def myfunc(a):
            if a == 0:
                raise error_class
            return 1

        f = implemented_function('f', myfunc)
        expr = f(val)
        assert expr == f(val)


def test_imps_wrong_args():
    pytest.raises(ValueError, lambda: implemented_function(sin, lambda x: x))


def test_lambdify_imps():
    # Test lambdify with implemented functions
    # first test basic (diofant) lambdify
    f = diofant.cos
    assert lambdify(x, f(x))(0) == 1
    assert lambdify(x, 1 + f(x))(0) == 2
    assert lambdify((x, y), y + f(x))(0, 1) == 2
    # make an implemented function and test
    f = implemented_function('f', lambda x: x + 100)
    assert lambdify(x, f(x))(0) == 100
    assert lambdify(x, 1 + f(x))(0) == 101
    assert lambdify((x, y), y + f(x))(0, 1) == 101
    # Can also handle tuples, lists, dicts as expressions
    lam = lambdify(x, (f(x), x))
    assert lam(3) == (103, 3)
    lam = lambdify(x, [f(x), x])
    assert lam(3) == [103, 3]
    lam = lambdify(x, [f(x), (f(x), x)])
    assert lam(3) == [103, (103, 3)]
    lam = lambdify(x, {f(x): x})
    assert lam(3) == {103: 3}
    lam = lambdify(x, {f(x): x})
    assert lam(3) == {103: 3}
    lam = lambdify(x, {x: f(x)})
    assert lam(3) == {3: 103}
    # Check that imp preferred to other namespaces by default
    d = {'f': lambda x: x + 99}
    lam = lambdify(x, f(x), d)
    assert lam(3) == 103
    # Unless flag passed
    lam = lambdify(x, f(x), d, use_imps=False)
    assert lam(3) == 102


def test_dummification():
    F = Function('F')
    G = Function('G')
    # "\alpha" is not a valid python variable name
    # lambdify should sub in a dummy for it, and return
    # without a syntax error
    alpha = symbols(r'\alpha')
    some_expr = 2 * F(t)**2 / G(t)
    lam = lambdify((F(t), G(t)), some_expr)
    assert lam(3, 9) == 2
    lam = lambdify(sin(t), 2 * sin(t)**2)
    assert lam(F(t)) == 2 * F(t)**2
    # Test that \alpha was properly dummified
    lam = lambdify((alpha, t), 2*alpha + t)
    assert lam(2, 1) == 5
    pytest.raises(SyntaxError, lambda: lambdify(F(t) * G(t), F(t) * G(t) + 5))
    pytest.raises(SyntaxError, lambda: lambdify(2 * F(t), 2 * F(t) + 5))
    pytest.raises(SyntaxError, lambda: lambdify(2 * F(t), 4 * F(t) + 5))


def test_python_keywords():
    # Test for issue sympy/sympy#7452. The automatic dummification should ensure use of
    # Python reserved keywords as symbol names will create valid lambda
    # functions. This is an additional regression test.
    python_if = symbols('if')
    expr = python_if / 2
    f = lambdify(python_if, expr)
    assert f(4.0) == 2.0


def test_lambdify_docstring():
    func = lambdify((w, x, y, z), w + x + y + z)
    assert func.__doc__ == (
        'Created with lambdify. Signature:\n\n'
        'func(w, x, y, z)\n\n'
        'Expression:\n\n'
        'w + x + y + z')
    syms = symbols('a1:26')
    func = lambdify(syms, sum(syms))
    assert func.__doc__ == (
        'Created with lambdify. Signature:\n\n'
        'func(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15,\n'
        '        a16, a17, a18, a19, a20, a21, a22, a23, a24, a25)\n\n'
        'Expression:\n\n'
        'a1 + a10 + a11 + a12 + a13 + a14 + a15 + a16 + a17 + a18 + a19 + a2 + a20 +...')


# ================= Test special printers ==========================


def test_special_printers():
    class IntervalPrinter(LambdaPrinter):
        """Use ``lambda`` printer but print numbers as ``mpi`` intervals."""

        def _print_Integer(self, expr):
            return f"mpi('{super()._print_Integer(expr)}')"

        def _print_Rational(self, expr):
            return f"mpi('{super()._print_Rational(expr)}')"

    def intervalrepr(expr):
        return IntervalPrinter().doprint(expr)

    expr = diofant.sqrt(diofant.sqrt(2) + diofant.sqrt(3)) + diofant.Rational(1, 2)

    func0 = lambdify((), expr, modules='mpmath', printer=intervalrepr)
    func1 = lambdify((), expr, modules='mpmath', printer=IntervalPrinter)
    func2 = lambdify((), expr, modules='mpmath', printer=IntervalPrinter())

    mpi = type(mpmath.mpi(1, 2))

    assert isinstance(func0(), mpi)
    assert isinstance(func1(), mpi)
    assert isinstance(func2(), mpi)


def test_true_false():
    # We want exact is comparison here, not just ==
    assert lambdify([], true)() is True
    assert lambdify([], false)() is False


def test_ITE():
    assert lambdify((x, y, z), ITE(x, y, z))(True, 5, 3) == 5
    assert lambdify((x, y, z), ITE(x, y, z))(False, 5, 3) == 3


def test_Min_Max():
    # see sympy/sympy#10375
    assert lambdify((x, y, z), Min(x, y, z))(1, 2, 3) == 1
    assert lambdify((x, y, z), Max(x, y, z))(1, 2, 3) == 3


def test_sympyissue_12092():
    f = implemented_function('f', lambda x: x**2)
    assert f(f(2)).evalf() == Float(16)


def test_sympyissue_23224():
    f = lambdify([], (1,))
    assert f() == (1,)



"""
Main Random Variables Module.

Defines abstract random variable type.
Contains interfaces for probability space object (PSpace) as well as standard
operators, P, E, sample, density, where

See Also
========
diofant.stats.crv
diofant.stats.frv
diofant.stats.rv_interface

"""

from __future__ import annotations

from ..abc import x
from ..core import (Add, Eq, Equality, Expr, Integer, Lambda, Symbol, Tuple,
                    oo, sympify)
from ..core.logic import fuzzy_or
from ..core.relational import Relational
from ..functions import DiracDelta
from ..logic.boolalg import Boolean, false, true
from ..sets import FiniteSet, ProductSet
from ..solvers import solve
from ..utilities import lambdify


class RandomDomain(Expr):
    """
    Represents a set of variables and the values which they can take

    See Also
    ========
    diofant.stats.crv.ContinuousDomain
    diofant.stats.frv.FiniteDomain

    """

    is_ProductDomain = False
    is_Finite = False
    is_Continuous = False

    def __new__(cls, symbols, *args):
        symbols = FiniteSet(*symbols)
        return Expr.__new__(cls, symbols, *args)

    @property
    def set(self):
        return self.args[1]

    def __contains__(self, other):
        raise NotImplementedError

    def integrate(self, expr):
        raise NotImplementedError


class SingleDomain(RandomDomain):
    """
    A single variable and its domain

    See Also
    ========

    diofant.stats.crv.SingleContinuousDomain
    diofant.stats.frv.SingleFiniteDomain

    """

    def __new__(cls, symbol, set):
        assert symbol.is_Symbol
        return Expr.__new__(cls, symbol, set)

    @property
    def symbol(self):
        return self.args[0]

    @property
    def symbols(self):
        return FiniteSet(self.symbol)


class ConditionalDomain(RandomDomain):
    """
    A RandomDomain with an attached condition

    See Also
    ========

    diofant.stats.crv.ConditionalContinuousDomain
    diofant.stats.frv.ConditionalFiniteDomain

    """

    def __new__(cls, fulldomain, condition):
        condition = condition.xreplace({rs: rs.symbol
                                        for rs in random_symbols(condition)})
        return Expr.__new__(cls, fulldomain, condition)

    @property
    def symbols(self):
        return self.fulldomain.symbols

    @property
    def fulldomain(self):
        return self.args[0]

    @property
    def condition(self):
        return self.args[1]

    @property
    def set(self):
        raise NotImplementedError('Set of Conditional Domain not Implemented')


class PSpace(Expr):
    """
    A Probability Space

    Probability Spaces encode processes that equal different values
    probabilistically. These underly Random Symbols which occur in Diofant
    expressions and contain the mechanics to evaluate statistical statements.

    See Also
    ========
    diofant.stats.crv.ContinuousPSpace
    diofant.stats.frv.FinitePSpace

    """

    is_Finite: bool | None = None
    is_Continuous: bool | None = None

    @property
    def values(self):
        return frozenset(RandomSymbol(self, sym) for sym in self.domain.symbols)

    @property
    def symbols(self):
        return self.domain.symbols

    def where(self, condition):
        raise NotImplementedError

    def compute_density(self, expr):
        raise NotImplementedError

    def sample(self):
        raise NotImplementedError

    def probability(self, condition):
        raise NotImplementedError

    def integrate(self, expr):
        raise NotImplementedError


class SinglePSpace(PSpace):
    """
    Represents the probabilities of a set of random events that can be
    attributed to a single variable/symbol.

    """

    def __new__(cls, s, distribution):
        if isinstance(s, str):
            s = Symbol(s)
        if not isinstance(s, Symbol):
            raise TypeError('s should have been string or Symbol')
        return Expr.__new__(cls, s, distribution)

    @property
    def value(self):
        return RandomSymbol(self, self.symbol)

    @property
    def symbol(self):
        return self.args[0]

    @property
    def distribution(self):
        return self.args[1]


class RandomSymbol(Expr):
    """
    Random Symbols represent ProbabilitySpaces in Diofant Expressions
    In principle they can take on any value that their symbol can take on
    within the associated PSpace with probability determined by the PSpace
    Density.

    Random Symbols contain pspace and symbol properties.
    The pspace property points to the represented Probability Space
    The symbol is a standard Diofant Symbol that is used in that probability space
    for example in defining a density.

    You can form normal Diofant expressions using RandomSymbols and operate on
    those expressions with the Functions

    E - Expectation of a random expression
    P - Probability of a condition
    density - Probability Density of an expression
    given - A new random expression (with new random symbols) given a condition

    An object of the RandomSymbol type should almost never be created by the
    user. They tend to be created instead by the PSpace class's value method.
    Traditionally a user doesn't even do this but instead calls one of the
    convenience functions Normal, Exponential, Coin, Die, FiniteRV, etc....

    """

    def __new__(cls, pspace, symbol):
        if not isinstance(symbol, Symbol):
            raise TypeError('symbol should be of type Symbol')
        if not isinstance(pspace, PSpace):
            raise TypeError('pspace variable should be of type PSpace')
        return Expr.__new__(cls, pspace, symbol)

    is_finite = True
    is_Symbol = True
    is_Atom = True

    _diff_wrt = True

    pspace = property(lambda self: self.args[0])
    symbol = property(lambda self: self.args[1])
    name = property(lambda self: self.symbol.name)

    def _eval_is_positive(self):
        return self.symbol.is_positive

    def _eval_is_integer(self):
        return self.symbol.is_integer

    def _eval_is_extended_real(self):
        return fuzzy_or([self.symbol.is_extended_real,
                         self.pspace.is_extended_real])

    def _eval_is_commutative(self):
        return self.symbol.is_commutative

    def _hashable_content(self):
        return self.pspace, self.symbol

    @property
    def free_symbols(self):
        return {self}


class ProductPSpace(PSpace):
    """
    A probability space resulting from the merger of two independent probability
    spaces.

    Often created using the function, pspace

    """

    def __new__(cls, *spaces):
        rs_space_dict = {}
        for space in spaces:
            for value in space.values:
                rs_space_dict[value] = space

        symbols = FiniteSet(*[val.symbol for val in rs_space_dict])

        # Overlapping symbols
        if len(symbols) < sum(len(space.symbols) for space in spaces):
            raise ValueError('Overlapping Random Variables')

        new_cls = cls
        if all(space.is_Finite for space in spaces):
            from .frv import ProductFinitePSpace
            new_cls = ProductFinitePSpace
        if all(space.is_Continuous for space in spaces):
            from .crv import ProductContinuousPSpace
            new_cls = ProductContinuousPSpace

        obj = Expr.__new__(new_cls, *FiniteSet(*spaces))

        return obj

    @property
    def rs_space_dict(self):
        d = {}
        for space in self.spaces:
            for value in space.values:
                d[value] = space
        return d

    @property
    def symbols(self):
        return FiniteSet(*[val.symbol for val in self.rs_space_dict])

    @property
    def spaces(self):
        return FiniteSet(*self.args)

    @property
    def values(self):
        return sumsets(space.values for space in self.spaces)

    def integrate(self, expr, rvs=None, **kwargs):
        rvs = rvs or self.values
        rvs = frozenset(rvs)
        for space in self.spaces:
            expr = space.integrate(expr, rvs & space.values, **kwargs)
        return expr

    @property
    def domain(self):
        return ProductDomain(*[space.domain for space in self.spaces])

    @property
    def density(self):
        raise NotImplementedError('Density not available for ProductSpaces')

    def sample(self):
        return {k: v for space in self.spaces
                for k, v in space.sample().items()}


class ProductDomain(RandomDomain):
    """
    A domain resulting from the merger of two independent domains

    See Also
    ========

    diofant.stats.crv.ProductContinuousDomain
    diofant.stats.frv.ProductFiniteDomain

    """

    is_ProductDomain = True

    def __new__(cls, *domains):
        # Flatten any product of products
        domains2 = []
        for domain in domains:
            if not domain.is_ProductDomain:
                domains2.append(domain)
            else:
                domains2.extend(domain.domains)
        domains2 = FiniteSet(*domains2)

        new_cls = cls
        if all(domain.is_Finite for domain in domains2):
            from .frv import ProductFiniteDomain
            new_cls = ProductFiniteDomain
        if all(domain.is_Continuous for domain in domains2):
            from .crv import ProductContinuousDomain
            new_cls = ProductContinuousDomain

        return Expr.__new__(new_cls, *domains2)

    @property
    def symbols(self):
        return FiniteSet(*[sym for domain in self.domains
                           for sym in domain.symbols])

    @property
    def domains(self):
        return self.args

    @property
    def set(self):
        return ProductSet(domain.set for domain in self.domains)

    def __contains__(self, other):
        # Split event into each subdomain
        for domain in self.domains:
            # Collect the parts of this event which associate to this domain
            elem = frozenset(item for item in other
                             if domain.symbols.contains(item[0]) == true)
            # Test this sub-event
            if elem not in domain:
                return False
        # All subevents passed
        return True


def random_symbols(expr):
    """Returns all RandomSymbols within a Diofant Expression."""
    try:
        return list(expr.atoms(RandomSymbol))
    except AttributeError:
        return []


def pspace(expr):
    """
    Returns the underlying Probability Space of a random expression.

    For internal use.

    Examples
    ========

    >>> from diofant.stats import Normal
    >>> X = Normal('X', 0, 1)
    >>> pspace(2*X + 1) == X.pspace
    True

    """
    expr = sympify(expr)
    rvs = random_symbols(expr)
    if not rvs:
        raise ValueError(f'Expression containing Random Variable expected, not {expr}')
    # If only one space present
    if all(rv.pspace == rvs[0].pspace for rv in rvs):
        return rvs[0].pspace
    # Otherwise make a product space
    return ProductPSpace(*[rv.pspace for rv in rvs])


def sumsets(sets):
    """Union of sets."""
    return frozenset().union(*sets)


def rs_swap(a, b):
    """
    Build a dictionary to swap RandomSymbols based on their underlying symbol.

    i.e.
    if    ``X = ('x', pspace1)``
    and   ``Y = ('x', pspace2)``
    then ``X`` and ``Y`` match and the key, value pair
    ``{X:Y}`` will appear in the result

    Inputs: collections a and b of random variables which share common symbols
    Output: dict mapping RVs in a to RVs in b

    """
    d = {}
    for rsa in a:
        d[rsa] = [rsb for rsb in b if rsa.symbol == rsb.symbol][0]
    return d


def given(expr, condition=None, **kwargs):
    r"""Conditional Random Expression.

    From a random expression and a condition on that expression creates a new
    probability space from the condition and returns the same expression on that
    conditional probability space.

    Examples
    ========

    >>> from diofant.stats import Die
    >>> X = Die('X', 6)
    >>> Y = given(X, X > 3)
    >>> density(Y).dict
    {4: 1/3, 5: 1/3, 6: 1/3}

    Following convention, if the condition is a random symbol then that symbol
    is considered fixed.

    >>> from diofant.stats import Normal

    >>> X = Normal('X', 0, 1)
    >>> Y = Normal('Y', 0, 1)
    >>> pprint(density(X + Y, Y)(z), use_unicode=False)
                    2
           -(-Y + z)
           -----------
      ___       2
    \/ 2 *E
    ------------------
             ____
         2*\/ pi

    """
    if not random_symbols(condition) or pspace_independent(expr, condition):
        return expr

    if isinstance(condition, RandomSymbol):
        condition = Eq(condition, condition.symbol)

    condsymbols = random_symbols(condition)
    if (isinstance(condition, Equality) and len(condsymbols) == 1 and
            not isinstance(pspace(expr).domain, ConditionalDomain)):
        rv = tuple(condsymbols)[0]
        results = solve(condition, rv)
        return sum(expr.subs(res) for res in results)

    # Get full probability space of both the expression and the condition
    fullspace = pspace(Tuple(expr, condition))
    # Build new space given the condition
    space = fullspace.conditional_space(condition, **kwargs)
    # Dictionary to swap out RandomSymbols in expr with new RandomSymbols
    # That point to the new conditional space
    swapdict = rs_swap(fullspace.values, space.values)
    # Swap random variables in the expression
    expr = expr.xreplace(swapdict)
    return expr


def expectation(expr, condition=None, numsamples=None, evaluate=True, **kwargs):
    """
    Returns the expected value of a random expression

    Parameters
    ==========

    expr : Expr containing RandomSymbols
        The expression of which you want to compute the expectation value
    given : Expr containing RandomSymbols
        A conditional expression. E(X, X>0) is expectation of X given X > 0
    numsamples : int
        Enables sampling and approximates the expectation with this many samples
    evalf : Bool (defaults to True)
        If sampling return a number rather than a complex expression
    evaluate : Bool (defaults to True)
        In case of continuous systems return unevaluated integral

    Examples
    ========

    >>> from diofant.stats import Die, E
    >>> X = Die('X', 6)
    >>> E(X)
    7/2
    >>> E(2*X + 1)
    8

    >>> E(X, X > 3)  # Expectation of X given that it is above 3
    5

    """
    if not random_symbols(expr):  # expr isn't random?
        return expr
    if numsamples:  # Computing by monte carlo sampling?
        return sampling_E(expr, condition, numsamples=numsamples, **kwargs)

    # Create new expr and recompute E
    if condition is not None:  # If there is a condition
        return expectation(given(expr, condition), evaluate=evaluate)

    # A few known statements for efficiency

    if expr.is_Add:  # We know that E is Linear
        return Add(*[expectation(arg, evaluate=evaluate)
                     for arg in expr.args])

    # Otherwise case is simple, pass work off to the ProbabilitySpace
    result = pspace(expr).integrate(expr)
    if evaluate and hasattr(result, 'doit'):
        return result.doit(**kwargs)
    else:
        return result


def probability(condition, given_condition=None, numsamples=None,
                evaluate=True, **kwargs):
    """
    Probability that a condition is true, optionally given a second condition

    Parameters
    ==========

    condition : Combination of Relationals containing RandomSymbols
        The condition of which you want to compute the probability
    given_condition : Combination of Relationals containing RandomSymbols
        A conditional expression. P(X > 1, X > 0) is expectation of X > 1
        given X > 0
    numsamples : int
        Enables sampling and approximates the probability with this many samples
    evaluate : Bool (defaults to True)
        In case of continuous systems return unevaluated integral

    Examples
    ========

    >>> from diofant.stats import Die, P
    >>> X, Y = Die('X', 6), Die('Y', 6)
    >>> P(X > 3)
    1/2
    >>> P(Eq(X, 5), X > 2)  # Probability that X == 5 given that X > 2
    1/4
    >>> P(X > Y)
    5/12

    """
    condition = sympify(condition)
    given_condition = sympify(given_condition)

    if given_condition is not None and \
            not isinstance(given_condition, (Relational, Boolean)):
        raise ValueError(f'{given_condition} is not a relational or combination of relationals')
    if given_condition == false:
        return Integer(0)
    if not isinstance(condition, (Relational, Boolean)):
        raise ValueError(f'{condition} is not a relational or combination of relationals')
    if condition == true:
        return Integer(1)
    if condition == false:
        return Integer(0)

    if numsamples:
        return sampling_P(condition, given_condition, numsamples=numsamples,
                          **kwargs)
    if given_condition is not None:  # If there is a condition
        # Recompute on new conditional expr
        return probability(given(condition, given_condition, **kwargs), **kwargs)

    # Otherwise pass work off to the ProbabilitySpace
    result = pspace(condition).probability(condition, **kwargs)
    if evaluate and hasattr(result, 'doit'):
        return result.doit()
    else:
        return result


class Density(Expr):
    """Probability density."""

    expr = property(lambda self: self.args[0])

    @property
    def condition(self):
        if len(self.args) > 1:
            return self.args[1]

    def doit(self, **kwargs):
        evaluate = kwargs.pop('evaluate', True)

        expr, condition = self.expr, self.condition
        if condition is not None:
            # Recompute on new conditional expr
            expr = given(expr, condition, **kwargs)
        if not random_symbols(expr):
            return Lambda(x, DiracDelta(x - expr))
        if (isinstance(expr, RandomSymbol) and
                hasattr(expr.pspace, 'distribution') and
                isinstance(pspace(expr), SinglePSpace)):
            return expr.pspace.distribution
        result = pspace(expr).compute_density(expr, **kwargs)

        if evaluate and hasattr(result, 'doit'):
            return result.doit()
        else:
            return result


def density(expr, condition=None, evaluate=True, numsamples=None, **kwargs):
    """
    Probability density of a random expression, optionally given a second
    condition.

    This density will take on different forms for different types of
    probability spaces. Discrete variables produce Dicts. Continuous
    variables produce Lambdas.

    Parameters
    ==========

    expr : Expr containing RandomSymbols
        The expression of which you want to compute the density value
    condition : Relational containing RandomSymbols
        A conditional expression. density(X > 1, X > 0) is density of X > 1
        given X > 0
    numsamples : int
        Enables sampling and approximates the density with this many samples

    Examples
    ========

    >>> from diofant.stats import Die, Normal

    >>> D = Die('D', 6)
    >>> X = Normal(x, 0, 1)

    >>> density(D).dict
    {1: 1/6, 2: 1/6, 3: 1/6, 4: 1/6, 5: 1/6, 6: 1/6}
    >>> density(2*D).dict
    {2: 1/6, 4: 1/6, 6: 1/6, 8: 1/6, 10: 1/6, 12: 1/6}
    >>> density(X)(x)
    sqrt(2)*E**(-x**2/2)/(2*sqrt(pi))

    """
    if numsamples:
        return sampling_density(expr, condition, numsamples=numsamples,
                                **kwargs)

    kwargs['evaluate'] = evaluate

    return Density(expr, condition).doit(**kwargs)


def cdf(expr, condition=None, evaluate=True, **kwargs):
    """
    Cumulative Distribution Function of a random expression.

    optionally given a second condition

    This density will take on different forms for different types of
    probability spaces.
    Discrete variables produce Dicts.
    Continuous variables produce Lambdas.

    Examples
    ========

    >>> from diofant.stats import Die, Normal

    >>> D = Die('D', 6)
    >>> X = Normal('X', 0, 1)

    >>> density(D).dict
    {1: 1/6, 2: 1/6, 3: 1/6, 4: 1/6, 5: 1/6, 6: 1/6}
    >>> cdf(D)
    {1: 1/6, 2: 1/3, 3: 1/2, 4: 2/3, 5: 5/6, 6: 1}
    >>> cdf(3*D, D > 2)
    {9: 1/4, 12: 1/2, 15: 3/4, 18: 1}

    >>> cdf(X)
    Lambda(_z, erf(sqrt(2)*_z/2)/2 + 1/2)

    """
    if condition is not None:  # If there is a condition
        # Recompute on new conditional expr
        return cdf(given(expr, condition, **kwargs), **kwargs)

    # Otherwise pass work off to the ProbabilitySpace
    result = pspace(expr).compute_cdf(expr, **kwargs)

    if evaluate and hasattr(result, 'doit'):
        return result.doit()
    else:
        return result


def where(condition, given_condition=None, **kwargs):
    """
    Returns the domain where a condition is True.

    Examples
    ========

    >>> from diofant.stats import Die, Normal

    >>> D1, D2 = Die('a', 6), Die('b', 6)
    >>> a, b = D1.symbol, D2.symbol
    >>> X = Normal('x', 0, 1)

    >>> where(X**2 < 1)
    Domain: (-1 < x) & (x < 1)

    >>> where(X**2 < 1).set
    (-1, 1)

    >>> where(And(D1 <= D2, D2 < 3))
    Domain: (Eq(a, 1) & Eq(b, 1)) | (Eq(a, 1) & Eq(b, 2)) | (Eq(a, 2) & Eq(b, 2))

    """
    if given_condition is not None:  # If there is a condition
        # Recompute on new conditional expr
        return where(given(condition, given_condition, **kwargs), **kwargs)

    # Otherwise pass work off to the ProbabilitySpace
    return pspace(condition).where(condition, **kwargs)


def sample(expr, condition=None, **kwargs):
    """
    A realization of the random expression

    Examples
    ========

    >>> from diofant.stats import Die
    >>> X, Y, Z = Die('X', 6), Die('Y', 6), Die('Z', 6)

    >>> die_roll = sample(X + Y + Z)  # A random realization of three dice

    """
    return next(sample_iter(expr, condition, numsamples=1))


def sample_iter(expr, condition=None, numsamples=oo, **kwargs):
    """
    Returns an iterator of realizations from the expression given a condition

    expr: Random expression to be realized
    condition: A conditional expression (optional)
    numsamples: Length of the iterator (defaults to infinity)

    Examples
    ========

    >>> from diofant.stats import Normal
    >>> X = Normal('X', 0, 1)
    >>> expr = X*X + 3
    >>> iterator = sample_iter(expr, numsamples=3)
    >>> list(iterator)  # doctest: +SKIP
    [12, 4, 7]

    See Also
    ========

    diofant.stats.sample
    diofant.stats.rv.sampling_P
    diofant.stats.rv.sampling_E

    """
    if condition is not None:
        ps = pspace(Tuple(expr, condition))
    else:
        ps = pspace(expr)

    rvs = list(ps.values)
    fn = lambdify(rvs, expr, **kwargs)
    if condition is not None:
        given_fn = lambdify(rvs, condition, **kwargs)

    # Check that lambdify can handle the expression
    # Some operations like Sum can prove difficult
    try:
        d = ps.sample()  # a dictionary that maps RVs to values
        args = [d[rv] for rv in rvs]
        fn(*args)
        if condition is not None:
            given_fn(*args)
    except (TypeError, ValueError) as exc:
        raise TypeError('Expr/condition too complex for lambdify') from exc

    def return_generator():
        count = 0
        while count < numsamples:
            d = ps.sample()  # a dictionary that maps RVs to values
            args = [d[rv] for rv in rvs]

            if condition is not None:  # Check that these values satisfy the condition
                gd = given_fn(*args)
                if gd not in (True, False):
                    raise ValueError(
                        'Conditions must not contain free symbols')
                if not gd:  # If the values don't satisfy then try again
                    continue

            yield fn(*args)
            count += 1
    return return_generator()


def sampling_P(condition, given_condition=None, numsamples=1,
               evalf=True, **kwargs):
    """
    Sampling version of P

    See Also
    ========
    diofant.stats.P
    diofant.stats.rv.sampling_E
    diofant.stats.rv.sampling_density

    """
    count_true = 0
    count_false = 0

    samples = sample_iter(condition, given_condition,
                          numsamples=numsamples, **kwargs)

    for x in samples:
        if x:
            count_true += 1
        else:
            count_false += 1

    result = Integer(count_true) / numsamples
    return result.evalf()


def sampling_E(expr, given_condition=None, numsamples=1,
               evalf=True, **kwargs):
    """
    Sampling version of E

    See Also
    ========
    diofant.stats.P
    diofant.stats.rv.sampling_P
    diofant.stats.rv.sampling_density

    """
    samples = sample_iter(expr, given_condition,
                          numsamples=numsamples, **kwargs)

    result = Add(*list(samples)) / numsamples
    return result.evalf(strict=False)


def sampling_density(expr, given_condition=None, numsamples=1, **kwargs):
    """
    Sampling version of density

    See Also
    ========
    diofant.stats.density
    diofant.stats.rv.sampling_P
    diofant.stats.rv.sampling_E

    """
    results = {}
    for result in sample_iter(expr, given_condition,
                              numsamples=numsamples, **kwargs):
        results[result] = results.get(result, 0) + 1
    return results


def dependent(a, b):
    """
    Dependence of two random expressions

    Two expressions are independent if knowledge of one does not change
    computations on the other.

    Examples
    ========

    >>> from diofant.stats import Normal

    >>> X, Y = Normal('X', 0, 1), Normal('Y', 0, 1)
    >>> dependent(X, Y)
    False
    >>> dependent(2*X + Y, -Y)
    True
    >>> X, Y = given(Tuple(X, Y), Eq(X + Y, 3))
    >>> dependent(X, Y)
    True

    See Also
    ========
    diofant.stats.rv.independent

    """
    if pspace_independent(a, b):
        return False

    z = Symbol('z', extended_real=True)
    # Dependent if density is unchanged when one is given information about
    # the other
    return (density(a, Eq(b, z)) != density(a) or
            density(b, Eq(a, z)) != density(b))


def independent(a, b):
    """
    Independence of two random expressions

    Two expressions are independent if knowledge of one does not change
    computations on the other.

    Examples
    ========

    >>> from diofant.stats import Normal

    >>> X, Y = Normal('X', 0, 1), Normal('Y', 0, 1)
    >>> independent(X, Y)
    True
    >>> independent(2*X + Y, -Y)
    False
    >>> X, Y = given(Tuple(X, Y), Eq(X + Y, 3))
    >>> independent(X, Y)
    False

    See Also
    ========
    diofant.stats.rv.dependent

    """
    return not dependent(a, b)


def pspace_independent(a, b):
    """
    Tests for independence between a and b by checking if their PSpaces have
    overlapping symbols. This is a sufficient but not necessary condition for
    independence and is intended to be used internally.

    Notes
    =====

    pspace_independent(a, b) implies independent(a, b)
    independent(a, b) does not imply pspace_independent(a, b)

    """
    a_symbols = set(pspace(b).symbols)
    b_symbols = set(pspace(a).symbols)

    if len(a_symbols.intersection(b_symbols)) == 0:
        return True


def rv_subs(expr):
    """Given a random expression replace all random variables with their symbols."""
    symbols = random_symbols(expr)
    if not symbols:
        return expr
    swapdict = {rv: rv.symbol for rv in symbols}
    return expr.xreplace(swapdict)


class NamedArgsMixin:
    """Helper class for named arguments."""

    _argnames: tuple[str, ...] = ()

    def __getattr__(self, attr):
        try:
            return self.args[self._argnames.index(attr)]
        except ValueError as exc:
            raise AttributeError(f"'{type(self).__name__}' object "
                                 f"has not attribute '{attr}'") from exc


def _value_check(condition, message):
    """
    Check a condition on input value.

    Raises ValueError with message if condition is not True

    """
    if condition != true:
        raise ValueError(message)



import pytest

from diofant.abc import x
from diofant.core.rules import Transform


__all__ = ()


def test_Transform():
    add1 = Transform(lambda x: x + 1, lambda x: x % 2 == 1)
    assert add1[1] == 2
    assert (1 in add1) is True
    assert add1.get(1) == 2

    pytest.raises(KeyError, lambda: add1[2])
    assert (2 in add1) is False
    assert add1.get(2) is None

    add1 = Transform(lambda x: x + 1)
    assert add1[1] == 2
    assert add1[x] == x + 1



"""
Implementation of DPLL algorithm.

Features:
  - Clause learning
  - Watch literal scheme
  - VSIDS heuristic

References
==========

* https://en.wikipedia.org/wiki/DPLL_algorithm

"""

from collections import defaultdict
from heapq import heappop, heappush

from ...utilities import default_sort_key, ordered
from ..boolalg import And, _find_predicates, to_cnf, to_int_repr


def dpll_satisfiable(expr, all_models=False):
    """
    Check satisfiability of a propositional sentence.

    Returns a model or a generator of all models if all_models is True.

    """
    clauses = And.make_args(to_cnf(expr))
    if False in clauses:
        if all_models:
            return (f for f in [False])
        return False
    symbols = sorted(_find_predicates(expr), key=default_sort_key)
    symbols_int_repr = range(1, len(symbols) + 1)
    clauses_int_repr = to_int_repr(clauses, symbols)

    solver = SATSolver(clauses_int_repr, symbols_int_repr, set(), symbols)
    models = solver._find_model()

    if all_models:
        return _all_models(models)

    try:
        return next(models)
    except StopIteration:
        return False


def _all_models(models):
    satisfiable = False
    try:
        while True:
            yield next(models)
            satisfiable = True
    except StopIteration:
        if not satisfiable:
            yield False


class SATSolver:
    """
    Class for representing a SAT solver capable of
     finding a model to a boolean theory in conjunctive
     normal form.

    """

    def __init__(self, clauses, variables, var_settings, symbols=None,
                 heuristic='vsids', clause_learning='none', INTERVAL=500):

        self.var_settings = var_settings
        self.heuristic = heuristic
        self.is_unsatisfied = False
        self._unit_prop_queue = []
        self.INTERVAL = INTERVAL

        if symbols is None:
            self.symbols = list(ordered(variables))
        else:
            self.symbols = symbols

        self._initialize_variables(variables)
        self._initialize_clauses(clauses)

        if 'vsids' == heuristic:
            self._vsids_init()
            self.heur_calculate = self._vsids_calculate
            self.heur_lit_assigned = self._vsids_lit_assigned
            self.heur_lit_unset = self._vsids_lit_unset
            self.heur_clause_added = self._vsids_clause_added

        else:
            raise NotImplementedError

        if 'none' == clause_learning:
            self.add_learned_clause = lambda x: None
            self.compute_conflict = lambda: None
        else:
            raise NotImplementedError

        # Create the base level
        self.levels = [Level(0)]
        self._current_level.varsettings = var_settings

        # Keep stats
        self.num_decisions = 0
        self.num_learned_clauses = 0
        self.original_num_clauses = len(self.clauses)

    def _initialize_variables(self, variables):
        """Set up the variable data structures needed."""
        self.sentinels = defaultdict(set)
        self.occurrence_count = defaultdict(int)
        self.variable_set = [False] * (len(variables) + 1)

    def _initialize_clauses(self, clauses):
        """
        Set up the clause data structures needed.

        For each clause, the following changes are made:
        - Unit clauses are queued for propagation right away.
        - Non-unit clauses have their first and last literals set as sentinels.
        - The number of clauses a literal appears in is computed.

        """
        self.clauses = []
        for cls in clauses:
            self.clauses.append(list(cls))

        for i, ci in enumerate(self.clauses):

            # Handle the unit clauses
            if 1 == len(ci):
                self._unit_prop_queue.append(ci[0])
                continue

            self.sentinels[ci[0]].add(i)
            self.sentinels[ci[-1]].add(i)

            for lit in ci:
                self.occurrence_count[lit] += 1

    def _find_model(self):
        """
        Main DPLL loop. Returns a generator of models.

        Variables are chosen successively, and assigned to be either
        True or False. If a solution is not found with this setting,
        the opposite is chosen and the search continues. The solver
        halts when every variable has a setting.

        """
        # We use this variable to keep track of if we should flip a
        # variable setting in successive rounds
        flip_var = False

        # Check if unit prop says the theory is unsat right off the bat
        self._simplify()
        if self.is_unsatisfied:
            return

        # While the theory still has clauses remaining
        while True:
            if flip_var:
                # We have just backtracked and we are trying to opposite literal
                flip_var = False
                lit = self._current_level.decision

            else:
                # Pick a literal to set
                lit = self.heur_calculate()
                self.num_decisions += 1

                # Stopping condition for a satisfying theory
                if 0 == lit:
                    yield {self.symbols[abs(lit) - 1]: lit > 0
                           for lit in self.var_settings}
                    while self._current_level.flipped:
                        self._undo()
                    if len(self.levels) == 1:
                        return
                    flip_lit = -self._current_level.decision
                    self._undo()
                    self.levels.append(Level(flip_lit, flipped=True))
                    flip_var = True
                    continue

                # Start the new decision level
                self.levels.append(Level(lit))

            # Assign the literal, updating the clauses it satisfies
            self._assign_literal(lit)

            # _simplify the theory
            self._simplify()

            # Check if we've made the theory unsat
            if self.is_unsatisfied:
                self.is_unsatisfied = False

                # We unroll all of the decisions until we can flip a literal
                while self._current_level.flipped:
                    self._undo()

                    # If we've unrolled all the way, the theory is unsat
                    if 1 == len(self.levels):
                        return

                # Detect and add a learned clause
                self.add_learned_clause(self.compute_conflict())

                # Try the opposite setting of the most recent decision
                flip_lit = -self._current_level.decision
                self._undo()
                self.levels.append(Level(flip_lit, flipped=True))
                flip_var = True

    ########################
    #    Helper Methods    #
    ########################

    @property
    def _current_level(self):
        """The current decision level data structure."""
        return self.levels[-1]

    def _clause_sat(self, cls):
        """Check if a clause is satisfied by the current variable setting."""
        for lit in self.clauses[cls]:
            if lit in self.var_settings:
                return True
        return False

    def _is_sentinel(self, lit, cls):
        """Check if a literal is a sentinel of a given clause."""
        return cls in self.sentinels[lit]

    def _assign_literal(self, lit):
        """Make a literal assignment.

        The literal assignment must be recorded as part of the current
        decision level. Additionally, if the literal is marked as a
        sentinel of any clause, then a new sentinel must be chosen. If
        this is not possible, then unit propagation is triggered and
        another literal is added to the queue to be set in the future.

        """
        self.var_settings.add(lit)
        self._current_level.var_settings.add(lit)
        self.variable_set[abs(lit)] = True
        self.heur_lit_assigned(lit)

        sentinel_list = list(self.sentinels[-lit])

        for cls in sentinel_list:
            if not self._clause_sat(cls):
                other_sentinel = None
                for newlit in self.clauses[cls]:
                    if newlit != -lit:
                        if self._is_sentinel(newlit, cls):
                            other_sentinel = newlit
                        elif not self.variable_set[abs(newlit)]:
                            self.sentinels[-lit].remove(cls)
                            self.sentinels[newlit].add(cls)
                            other_sentinel = None
                            break

                # Check if no sentinel update exists
                if other_sentinel:
                    self._unit_prop_queue.append(other_sentinel)

    def _undo(self):
        """Undo the changes of the most recent decision level."""
        # Undo the variable settings
        for lit in self._current_level.var_settings:
            self.var_settings.remove(lit)
            self.heur_lit_unset(lit)
            self.variable_set[abs(lit)] = False

        # Pop the level off the stack
        self.levels.pop()

    #########################
    #      Propagation      #
    #########################

    # Propagation methods should attempt to soundly simplify the boolean
    # theory, and return True if any simplification occurred and False
    # otherwise.

    def _simplify(self):
        """Iterate over the various forms of propagation to simplify the theory."""
        changed = True
        while changed:
            changed = False
            changed |= self._unit_prop()
            changed |= self._pure_literal()

    def _unit_prop(self):
        """Perform unit propagation on the current theory."""
        result = len(self._unit_prop_queue) > 0
        while self._unit_prop_queue:
            next_lit = self._unit_prop_queue.pop()
            if -next_lit in self.var_settings:
                self.is_unsatisfied = True
                self._unit_prop_queue = []
                return False
            else:
                self._assign_literal(next_lit)

        return result

    def _pure_literal(self):
        """Look for pure literals and assign them when found."""
        return False

    #########################
    #      Heuristics       #
    #########################
    def _vsids_init(self):
        """Initialize the data structures needed for the VSIDS heuristic."""
        self.lit_heap = []
        self.lit_scores = {}

        for var in range(1, len(self.variable_set)):
            self.lit_scores[var] = float(-self.occurrence_count[var])
            self.lit_scores[-var] = float(-self.occurrence_count[-var])
            heappush(self.lit_heap, (self.lit_scores[var], var))
            heappush(self.lit_heap, (self.lit_scores[-var], -var))

    def _vsids_decay(self):
        """Decay the VSIDS scores for every literal."""
        # We divide every literal score by 2 for a decay factor
        #  Note: This doesn't change the heap property
        for lit in self.lit_scores:
            self.lit_scores[lit] /= 2.0

    def _vsids_calculate(self):
        """VSIDS Heuristic Calculation."""
        if len(self.lit_heap) == 0:
            return 0

        # Clean out the front of the heap as long the variables are set
        while self.variable_set[abs(self.lit_heap[0][1])]:
            heappop(self.lit_heap)
            if len(self.lit_heap) == 0:
                return 0

        return heappop(self.lit_heap)[1]

    def _vsids_lit_assigned(self, lit):
        """Handle the assignment of a literal for the VSIDS heuristic."""

    def _vsids_lit_unset(self, lit):
        """Handle the unsetting of a literal for the VSIDS heuristic."""
        var = abs(lit)
        heappush(self.lit_heap, (self.lit_scores[var], var))
        heappush(self.lit_heap, (self.lit_scores[-var], -var))

    def _vsids_clause_added(self, cls):
        """Handle the addition of a new clause for the VSIDS heuristic."""
        self.num_learned_clauses += 1
        for lit in cls:
            self.lit_scores[lit] += 1

    ########################
    #   Clause Learning    #
    ########################

    def _simple_add_learned_clause(self, cls):
        """Add a new clause to the theory."""
        cls_num = len(self.clauses)
        self.clauses.append(cls)

        for lit in cls:
            self.occurrence_count[lit] += 1

        self.sentinels[cls[0]].add(cls_num)
        self.sentinels[cls[-1]].add(cls_num)

        self.heur_clause_added(cls)

    def _simple_compute_conflict(self):
        """Build a clause representing the fact that at least one decision made so far is wrong."""
        return [-(level.decision) for level in self.levels[1:]]


class Level:
    """
    Represents a single level in the DPLL algorithm, and contains
    enough information for a sound backtracking procedure.

    """

    def __init__(self, decision, flipped=False):
        self.decision = decision
        self.var_settings = set()
        self.flipped = flipped
